"""  This script provide the required functionality for consolidating the QANom annotation from multiple generators
 (typically two) into a final, non-redundant annotation. """
from typing import *

import pandas as pd

from qanom import utils
from qanom.annotations import common, decode_encode_answers
from qanom.annotations.decode_encode_answers \
    import Argument, Role, Response, encode_response, arg_length
from qanom.evaluation.metrics import iou

FINAL_COLUMNS = ['qasrl_id', 'sentence', 'target_idx', 'key', 'verb',
                 'worker_id', 'assign_id',
                 'is_verbal', 'verb_form',
                 'question', 'answer_range', 'answer',
                 'wh', 'subj', 'obj', 'obj2', 'aux', 'prep', 'verb_prefix',
                 'is_passive', 'is_negated']
# original annot files include also 'source_assign_id' and 'is_redundant' (for arbitration\validation)

def auto_consolidate_gen_annot_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return consolidated (non-redundant) QANom annotations.
    Algorithm is precision-oriented:
        take conjunction of is-verbal,
        filter only roles (questions) aligned by their answers,
        for answers: take larger set of answers;
            if even - take longer (and select the question corresponding to selected answers).
    :param df: generation annotation DataFrame containing multiple workers' annotation per predicate
    :return: annotation csv (encoded, not decoded) of the consolidated annotations
    """
    from annotations.decode_encode_answers import SPAN_SEPARATOR
    data_columns = ['qasrl_id', 'sentence', 'target_idx', 'key', 'verb', 'verb_form']
    to_be_conjoined_columns = ['worker_id', 'assign_id']
    """ Rest of columns are annotation-columns - they are to be decoded from consolidated Response.
        (Except from 'answer' which requires both answer_range from Response and sentence.) """
    pred_dfs: List[pd.DataFrame] = []
    for key, pred_df in df.groupby('key'):
        responses = {worker: decode_encode_answers.decode_response(worker_df)
                     for worker, worker_df in pred_df.groupby('worker_id')}
        if len(responses)<2:
            # only one generator for predicate
            print(f"Warning: predicate {key} has only one generator. Taking his response as final.")
            consolidated_response = list(responses.values())[0]
        else:
            # two generators = automatically consolidate responses
            consolidated_response = auto_consolidate_predicate(list(responses.values()))
        cons_pred_df = encode_response(consolidated_response, pred_df.sentence.to_list()[0])
        # add other columns to resulting df
        cons_pred_df = cons_pred_df.assign(**{col: pred_df[col].to_list()[0] for col in data_columns})
        cons_pred_df = cons_pred_df.assign(**{col: SPAN_SEPARATOR.join(set(pred_df[col])) for col in to_be_conjoined_columns})
        # re-order columns for convenience
        cons_pred_df = cons_pred_df[FINAL_COLUMNS]
        pred_dfs.append(cons_pred_df)
    out = pd.concat(pred_dfs)
    return out


def get_aligned_roles(resp1: Response, resp2: Response) -> List[Role]:
    # take only aligned arguments along with their full Role (=QA)
    from evaluation.alignment import find_matches
    alignment: Dict[Argument, Argument] = find_matches(resp1.all_args(), resp2.all_args())

    # Helper func:
    def orig_role(arg: Argument, response: Response) -> Role:
        """ get original role (question & argument-set) of arg. """
        roles_containing_arg = list(filter(lambda r: arg in r.arguments, response.roles))
        # shouldn't ever happen due to UI restriction on overlapping arguments
        assert len(
            roles_containing_arg) == 1, f"More or less than one role contains this arg: {arg}, roles: {roles_containing_arg}"
        return roles_containing_arg[0]

    def is_conflicting(role: Role, pre_selected_roles: Set[Role]) -> bool:
        """ Return True if any of role.arguments if conflicting (overlapping) with the pre-selected arguments. """
        overlapping = lambda a1, a2: iou(a1, a2) > 0
        return any(overlapping(arg, pre_arg)
                   for arg in role.arguments
                   for pre_role in pre_selected_roles
                   for pre_arg in pre_role.arguments)

    selected_roles: Set[Role] = set()

    def get_better_role(role1: Role, role2: Role) -> Union[Role, None]:
        """ Choose which role to select for consolidated response (from two aligned roles) """
        # Criteria 0: role doesn't conflict with roles we have previosuly selected
        role1_conflict = is_conflicting(role1, selected_roles)
        role2_conflict = is_conflicting(role2, selected_roles)
        if role1_conflict and role2_conflict:
            return None
        elif role1_conflict:
            return role2
        elif role2_conflict:
            return role1
        # Criteria 1: num of spans in argument-set
        if len(role1.arguments) != len(role2.arguments):
            return role1 if len(role1.arguments) > len(role2.arguments) else role2
        # Criteria 2: (overall) length of arguments
        num_tokens1 = sum(arg_length(arg) for arg in role1.arguments)
        num_tokens2 = sum(arg_length(arg) for arg in role2.arguments)
        if num_tokens1 != num_tokens2:
            return role1 if num_tokens1 > num_tokens2 else role2
        # no other criteria - arbitrarily select
        return role1

    for arg_w1, arg_w2 in alignment.items():
        # choose which role to select for consolidated response
        role1 = orig_role(arg_w1, resp1)
        role2 = orig_role(arg_w2, resp2)
        new_role = get_better_role(role1, role2)
        if new_role:
            selected_roles.add(new_role)
    return list(selected_roles)


def auto_consolidate_predicate(responses: List[Response], method: str = "intersection") -> Response:
    """
    :param responses:
    :param method: "intersection" | "majority"
    :return:
    """
    assert method in ["intersection", "majority"], "'method' must be 'intersection' or 'majority'!"
    if len(responses) == 1:
        return responses[0]
    assert len(responses) > 1, "Got 0 responses to consolidate"

    verb_form = responses[0].verb_form
    is_verbal_decisions = list(map(lambda r: r.is_verbal, responses))
    majority_is_verbal = utils.majority(is_verbal_decisions, whenTie=False)
    has_no_roles = list(map(lambda r: not bool(r.roles), responses))
    # in case no generator produced QAs for this predicate
    if all(has_no_roles):
        return Response(majority_is_verbal, verb_form)

    if len(responses) > 2:
        if method=="intersection":
            # recursive solution - taking only intersection roles
            cons_response1 = auto_consolidate_predicate(responses[:2])
            cons_response2 = auto_consolidate_predicate(responses[2:])
            return auto_consolidate_predicate([cons_response1, cons_response2], method)
        elif method=="majority":
            # todo
            raise NotImplementedError

    # assuming Iterable of 2 responses
    elif len(responses) == 2:
        selected_roles = get_aligned_roles(*responses[:2])
        return Response(majority_is_verbal, verb_form, selected_roles)


def auto_consolidation_iaa_experiment(dup_annot_df: pd.DataFrame) -> float:
    # returns argument F1 agreement
    import evaluation.evaluate_inter_annotator as eia
    desired_workers = ['A1FS8SBR4SDWYG', 'A21LONLNBOB8Q', 'A2A4UAFZ5LW71K', 'AJQGWGESKQT4Y']
    only4w = dup_annot_df[dup_annot_df.worker_id.isin(desired_workers)]
    df = only4w
    df['n_workers'] = df.groupby('key').worker_id.transform(pd.Series.nunique)
    only4w = df[df.n_workers == 4]
    print(f"number of predicates with full annotation of same 4 workers: {common.get_n_predicates(only4w)}")

    grp1, grp2 = ['A1FS8SBR4SDWYG', 'AJQGWGESKQT4Y'], ['A21LONLNBOB8Q', 'A2A4UAFZ5LW71K']
    r2grp = lambda r: "grp1" if r.worker_id in grp1 else "grp2"
    only4w['group'] = only4w.apply(r2grp, axis=1)
    grp2df = dict(list(only4w.groupby('group')))
    # generate consolidated df (automatic) for each group
    cons1_df = auto_consolidate_gen_annot_df(grp2df["grp1"])
    cons2_df = auto_consolidate_gen_annot_df(grp2df["grp2"])
    meta_df = pd.concat([cons1_df, cons2_df], ignore_index=True, sort=False)
    import annotations.decode_encode_answers as decode_encode
    meta_decoded_df = decode_encode.decode_qasrl(meta_df)
    # print IAA
    return eia.evaluate_inter_generator_agreement(meta_decoded_df)



""" Utils for candidate_extraction module """

def levenshteinDistance(s1, s2):
    if len(s1) > len(s2):
        s1, s2 = s2, s1

    distances = range(len(s1) + 1)
    for i2, c2 in enumerate(s2):
        distances_ = [i2+1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]


def results_by_edit_distance(orig_word, optional_results):
    # Build the result in the form of a list containing tuples (word, probability)
    result = [(w, levenshteinDistance(w, orig_word)) for w in set(optional_results)]
    result.sort(key=lambda w:len(w[0])) # tie breaker - prefer shorter verbs
    result.sort(key=lambda w:-w[1], reverse = True)

    # return all the possibilities sorted by probability
    return result



from typing import Optional
import json
from argparse import Namespace
from pathlib import Path
from transformers import Text2TextGenerationPipeline, AutoModelForSeq2SeqLM, AutoTokenizer

def get_markers_for_model(is_t5_model: bool) -> Namespace:
    special_tokens_constants = Namespace() 
    if is_t5_model:
        # T5 model have 100 special tokens by default
        special_tokens_constants.separator_input_question_predicate = "<extra_id_1>"
        special_tokens_constants.separator_output_answers = "<extra_id_3>"
        special_tokens_constants.separator_output_questions = "<extra_id_5>"  # if using only questions
        special_tokens_constants.separator_output_question_answer = "<extra_id_7>"
        special_tokens_constants.separator_output_pairs = "<extra_id_9>"
        special_tokens_constants.predicate_generic_marker = "<extra_id_10>" 
        special_tokens_constants.predicate_verb_marker = "<extra_id_11>" 
        special_tokens_constants.predicate_nominalization_marker = "<extra_id_12>" 

    else:
        special_tokens_constants.separator_input_question_predicate = "<question_predicate_sep>"
        special_tokens_constants.separator_output_answers = "<answers_sep>"
        special_tokens_constants.separator_output_questions = "<question_sep>"  # if using only questions
        special_tokens_constants.separator_output_question_answer = "<question_answer_sep>"
        special_tokens_constants.separator_output_pairs = "<qa_pairs_sep>"
        special_tokens_constants.predicate_generic_marker = "<predicate_marker>" 
        special_tokens_constants.predicate_verb_marker = "<verbal_predicate_marker>" 
        special_tokens_constants.predicate_nominalization_marker = "<nominalization_predicate_marker>" 
    return special_tokens_constants

def load_trained_model(name_or_path):
    import huggingface_hub as HFhub
    tokenizer = AutoTokenizer.from_pretrained(name_or_path)
    model = AutoModelForSeq2SeqLM.from_pretrained(name_or_path)  
    # load preprocessing_kwargs from the model repo on HF hub, or from the local model directory
    kwargs_filename = None
    if name_or_path.startswith("kleinay/") and 'preprocessing_kwargs.json' in HFhub.list_repo_files(name_or_path):
        kwargs_filename = HFhub.hf_hub_download(repo_id=name_or_path, filename="preprocessing_kwargs.json")
    elif Path(name_or_path).is_dir() and (Path(name_or_path) / "experiment_kwargs.json").exists():
        kwargs_filename = Path(name_or_path) / "experiment_kwargs.json"
    
    if kwargs_filename:
        preprocessing_kwargs = json.load(open(kwargs_filename)) 
        # integrate into model.config (for decoding args, e.g. "num_beams"), and save also as standalone object for preprocessing
        model.config.preprocessing_kwargs = Namespace(**preprocessing_kwargs)
        model.config.update(preprocessing_kwargs)
    return model, tokenizer


class QASRL_Pipeline(Text2TextGenerationPipeline):
    def __init__(self, model_repo: str, **kwargs):
        model, tokenizer = load_trained_model(model_repo)
        super().__init__(model, tokenizer, framework="pt")
        self.is_t5_model = "t5" in model.config.model_type
        self.special_tokens = get_markers_for_model(self.is_t5_model)
        # self.preprocessor = preprocessing.Preprocessor(model.config.preprocessing_kwargs, self.special_tokens)
        self.data_args = model.config.preprocessing_kwargs 
        # backward compatibility - default keyword values implemeted in `run_summarization`, thus not saved in `preprocessing_kwargs`
        if "predicate_marker_type" not in vars(self.data_args):
            self.data_args.predicate_marker_type = "generic"
        if "use_bilateral_predicate_marker" not in vars(self.data_args):
            self.data_args.use_bilateral_predicate_marker = True
        if "append_verb_form" not in vars(self.data_args):
            self.data_args.append_verb_form = True
        self._update_config(**kwargs)
    
    def _update_config(self, **kwargs):
        " Update self.model.config with initialization parameters and necessary defaults. "
        # set default values that will always override model.config, but can overriden by __init__ kwargs
        kwargs["max_length"] = kwargs.get("max_length", 80)
        # override model.config with kwargs
        for k,v in kwargs.items():
            self.model.config.__dict__[k] = v           
        
    def _sanitize_parameters(self, **kwargs):
        preprocess_kwargs, forward_kwargs, postprocess_kwargs = {}, {}, {} # super()._sanitize_parameters(**kwargs)
        forward_kwargs.update(kwargs.get("generate_kwargs", dict()))
        forward_kwargs.update(kwargs.get("model_kwargs", dict()))
        preprocess_keywords = ("predicate_marker", "predicate_type", "verb_form")
        for key in preprocess_keywords:
            if key in kwargs:
                preprocess_kwargs[key] = kwargs[key]

        return preprocess_kwargs, forward_kwargs, postprocess_kwargs

    def preprocess(self, inputs, predicate_marker="<predicate>", predicate_type=None, verb_form=None):
        # Here, inputs is string or list of strings; apply string postprocessing
        if isinstance(inputs, str):
            processed_inputs = self._preprocess_string(inputs, predicate_marker, predicate_type, verb_form)
        elif hasattr(inputs, "__iter__"):
            processed_inputs = [self._preprocess_string(s, predicate_marker, predicate_type, verb_form) for s in inputs]
        else:
            raise ValueError("inputs must be str or Iterable[str]")
        # Now pass to super.preprocess for tokenization
        return super().preprocess(processed_inputs)
    
    def _preprocess_string(self, seq: str, predicate_marker: str, predicate_type: Optional[str], verb_form: Optional[str]) -> str:
        sent_tokens = seq.split(" ")
        assert predicate_marker in sent_tokens, f"Input sentence must include a predicate-marker token ('{predicate_marker}') before the target predicate word"
        predicate_idx = sent_tokens.index(predicate_marker)
        sent_tokens.remove(predicate_marker)
        sentence_before_predicate = " ".join([sent_tokens[i] for i in range(predicate_idx)])
        predicate = sent_tokens[predicate_idx]
        sentence_after_predicate = " ".join([sent_tokens[i] for i in range(predicate_idx+1, len(sent_tokens))])
        
        if self.data_args.predicate_marker_type == "generic":
            predicate_marker = self.special_tokens.predicate_generic_marker    
        #  In case we want special marker for each predicate type: """
        elif self.data_args.predicate_marker_type == "pred_type":
            assert predicate_type is not None, "For this model, you must provide the `predicate_type` either when initializing QASRL_Pipeline(...) or when applying __call__(...) on it"
            assert predicate_type in ("verbal", "nominal"), f"`predicate_type` must be either 'verbal' or 'nominal'; got '{predicate_type}'"
            predicate_marker = {"verbal": self.special_tokens.predicate_verb_marker , 
                                "nominal": self.special_tokens.predicate_nominalization_marker 
                                }[predicate_type]

        if self.data_args.use_bilateral_predicate_marker:
            seq = f"{sentence_before_predicate} {predicate_marker} {predicate} {predicate_marker} {sentence_after_predicate}"
        else:
            seq = f"{sentence_before_predicate} {predicate_marker} {predicate} {sentence_after_predicate}"

        # embed also verb_form
        if self.data_args.append_verb_form and verb_form is None:
            raise ValueError(f"For this model, you must provide the `verb_form` of the predicate when applying __call__(...)")
        elif self.data_args.append_verb_form:
            seq = f"{seq} {self.special_tokens.separator_input_question_predicate} {verb_form} "
        else:
            seq = f"{seq} "
    
        # append source prefix (for t5 models)
        prefix = self._get_source_prefix(predicate_type)
        
        return prefix + seq
    
    def _get_source_prefix(self, predicate_type: Optional[str]):
        if not self.is_t5_model or self.data_args.source_prefix is None:
            return ''
        if "<predicate_type>" in self.data_args.source_prefix:
            if predicate_type is None:
                raise ValueError("source_prefix includes '<predicate_type>' but input has no `predicate_type`.")
            if self.data_args.source_prefix == "Generate QAs for <predicate_type> QASRL: ": # backwrad compatibility - "Generate QAs for <predicate_type> QASRL: " alone was a sign for a longer prefix 
                return f"Generate QAs for {predicate_type} QASRL: "
            else:
                return self.data_args.source_prefix.replace("Generate QAs for <predicate_type> QASRL: ", predicate_type)
        else:
            return self.data_args.source_prefix
        
    
    def _forward(self, *args, **kwargs):
        outputs = super()._forward(*args, **kwargs)
        return outputs


    def postprocess(self, model_outputs):
        output_seq = self.tokenizer.decode(
            model_outputs["output_ids"][0],
            skip_special_tokens=False,
            clean_up_tokenization_spaces=False,
        )
        output_seq = self._strip_output_seq(output_seq)
        qa_subseqs = output_seq.split(self.special_tokens.separator_output_pairs)
        qas = [self._postrocess_qa(qa_subseq) for qa_subseq in qa_subseqs]
        return {"generated_text": output_seq,
                "QAs": qas}
        
    def _strip_output_seq(self, output_seq: str) -> str:
        if output_seq.startswith(self.tokenizer.pad_token):
            output_seq = output_seq[len(self.tokenizer.pad_token):]
        if output_seq.endswith(self.tokenizer.pad_token):
            output_seq = output_seq.rstrip(self.tokenizer.pad_token)
        if output_seq.endswith(self.tokenizer.eos_token):
            output_seq = output_seq[:-len(self.tokenizer.eos_token)]
        elif output_seq.endswith(self.tokenizer.eos_token[:-1]):    
                # because previous pad_token-strip removed ">" from eos_token
            output_seq = output_seq[:-len(self.tokenizer.eos_token[:-1])]
        return output_seq.strip()
            
        
            
    def _postrocess_qa(self, seq: str) -> str:
        # split question and answers
        if self.special_tokens.separator_output_question_answer in seq:
            question, answer = seq.split(self.special_tokens.separator_output_question_answer)[:2]
        else:
            print("invalid format: no separator between question and answer found...")
            return None
            # question, answer = seq, '' # Or: backoff to only question  
        # skip "_" slots in questions
        question = ' '.join(t for t in question.split(' ') if t != '_')
        answers = [a.strip() for a in answer.split(self.special_tokens.separator_output_answers)]
        return {"question": question, "answers": answers}
    
    
if __name__ == "__main__":
    pipe = QASRL_Pipeline("kleinay/qanom-seq2seq-model-baseline")
    res1 = pipe("The student was interested in Luke 's <predicate> research about see animals .", verb_form="research", predicate_type="nominal")
    res2 = pipe(["The doctor was interested in Luke 's <predicate> treatment .",
                 "The Veterinary student was interested in Luke 's <predicate> treatment of sea animals ."], verb_form="treat", predicate_type="nominal", num_beams=10)
    res3 = pipe("A number of professions have <predicate> developed that specialize in the treatment of mental disorders .", verb_form="develop", predicate_type="verbal")
    print(res1)
    print(res2)
    print(res3)
    


from itertools import product
from typing import List, Dict

from qanom.annotations.decode_encode_answers import Argument, Role
from qanom.evaluation.metrics import iou, MATCH_IOU_THRESHOLD


def find_matches(sys_args: List[Argument], grt_args: List[Argument]) -> Dict[Argument, Argument]:
    """ Matching sys and grt arguments based in span overlap (IOU above threshold). """
    matches = ((sys_arg, grt_arg, iou(sys_arg, grt_arg))
               for sys_arg, grt_arg in product(sys_args, grt_args))
    # get any overlapping pairs between SYS and GRT
    matches = [(sys_arg, grt_arg, score)
               for (sys_arg, grt_arg, score) in matches
               if score > MATCH_IOU_THRESHOLD]
    matches = sorted(matches, key=lambda m: m[2], reverse=True)
    if not matches:
        return dict()

    used_sys_args, used_grt_args = set(), set()
    sys_to_grt = {}
    for (sys_arg, grt_arg, score) in matches:
        if grt_arg in used_grt_args or sys_arg in used_sys_args:
            continue
        sys_to_grt[sys_arg] = grt_arg
        used_sys_args.add(sys_arg)
        used_grt_args.add(grt_arg)

    return sys_to_grt


class RoleAlignment:

    def __init__(self, grt_roles: List[Role], sys_roles: List[Role]):
        self.grt_roles = grt_roles
        self.sys_roles = sys_roles
        self.sys_to_grt = {role: set() for role in sys_roles}
        self.grt_to_sys = {role: set() for role in grt_roles}

    def add_alignment(self, grt_role: Role, sys_role: Role):
        self.sys_to_grt[sys_role].add(grt_role)
        self.grt_to_sys[grt_role].add(sys_role)

    def has_single_alignment(self, role: Role, is_grt: bool):
        role_dict = self.grt_to_sys if is_grt else self.sys_to_grt
        n_aligned = len(role_dict[role])
        return n_aligned == 1


def find_role(arg: Argument, roles: List[Role]) -> Role:
    return next(role for role in roles if arg in role.arguments)


def align_by_argument(grt_roles: List[Role], sys_roles: List[Role],
                      sys_to_grt: Dict[Argument, Argument]) -> RoleAlignment:
    alignment = RoleAlignment(grt_roles, sys_roles)
    for sys_role in sys_roles:
        for sys_arg in sys_role.arguments:
            grt_arg = sys_to_grt.get(sys_arg)
            if not grt_arg:
                continue
            grt_role = find_role(grt_arg, grt_roles)
            alignment.add_alignment(grt_role, sys_role)
    return alignment


"""
Use lexical resources to extract nouns that are candidates for being deverbal nominalizations.

The script can be used for generating a JSON file that serves as input to the qasrl-crowdsourcing MTurk application
for annotating verbal arguments of nominalizations (qanom-unified branch).
It can also be used for generating a CSV files that serves as input to the predicate_detector.
"""

import argparse

if __name__ == "__main__":
    ap = argparse.ArgumentParser(description="Use lexical resources to extract nouns that are " +
                                                 "candidates for being deverbal nominalizations.")
    ap.add_argument('sentences_fn', type=str)
    ap.add_argument('output_fn', type=str)
    ap.add_argument('--read', dest='input_format', choices=['csv', 'jsonl', 'raw'], default='csv',
                    help="Define the format of sentences_fn to read from. \n"
                        +"csv is expecting a 'sentence' column and a 'sentence_id|sentenceId|qasrl_id' column.\n"
                        +"jsonl correspond to AllenNLP predictor's format, where each line is {'sentence': string}.\n"
                        +"raw is a text file, where each sentence is in a new line.")

    ap.add_argument('--write', dest='output_format', choices=['csv', 'json'], default='json',
                    help="Define the output format of candidate information. \n"
                             + "csv is the QANom default format. This is the format which predicate-detector model expects as input. \n"
                             + "json is used as input in the qasrl-crowdsourcing system when crowdsourcing QANom annotations.")
    # which resources to use - by default, use all three
    ap.add_argument('--no-wordnet', dest='wordnet', action='store_false')
    ap.add_argument('--no-catvar', dest='catvar', action='store_false')
    ap.add_argument('--no-affixes', dest='affixes_heuristic', action='store_false')

    args = ap.parse_args()

    from qanom.candidate_extraction import candidate_extraction
    candidate_extraction.main(args)



"""
A script with a set of post-processing utilities for the collected annotation files (CSVs).
"""
import json
import os.path
from typing import *

import pandas as pd

from qanom import utils
from qanom.annotations import common
from qanom.annotations.common import read_annot_csv, save_annot_csv, read_dir_of_csv


# generic post-processing function
def postprocess_annotation_files(orig_dir: str, dest_dir: str,
                                 process_annot_func: Callable[[pd.DataFrame,], pd.DataFrame],
                                 file_name_modification_func: Callable[[str,], str] = lambda s:s) -> NoReturn:
    """
    :param orig_dir: Directory from which to take the annottion to process (input)
    :param dest_dir: Directory to which the processed annotation files are to be exported
    :param process_annot_func: a function that gets an annot_df and returns a processed (i.e. corrected or changed,
    to some aspect) annot_df
    :param file_name_modification_func: how to change an annotation file-name from source-dir to dest-dir
    :return:
    """
    ann_files = [os.path.join(orig_dir, fn) for fn in os.listdir(orig_dir) if fn.endswith(".csv")]
    for orig_fn in ann_files:
        orig_df = read_annot_csv(orig_fn)
        new_df = process_annot_func(orig_df)
        # now export to file with same naming as orig (but in destination folder)
        orig_dir, orig_name = os.path.split(orig_fn)
        new_name = file_name_modification_func(orig_name)
        dest_fn = os.path.join(dest_dir, new_name)
        save_annot_csv(new_df, dest_fn)
        print(f"exported annotations to {dest_fn}")


def merge_csvs(csvs: List[str], dest: str) -> pd.DataFrame:
    dfs = [common.read_csv(csv_fn) for csv_fn in csvs]
    merged_df = pd.concat(dfs, ignore_index=True, sort=False)
    merged_df.to_csv(dest, index=False, encoding="utf-8")
    print(f"exported DataFrame with shape {merged_df.shape} to {dest}")
    return merged_df


def find_invalid_prompts(annot_df : pd.DataFrame) -> pd.DataFrame:
    """
    Verify that all the HITs of the annotated data are valid according to the latest version of
    candidate_extraction.py, where I have found and corrected errors about:
        [* Wordnet edit-distance threshold] (deprecated after next change)
        * Wordnet verbalize() algorithm - using lemmas instead of synsets (large impact!)
        * Tokenization issues - pos_tagger imposes different tokenization from sentence.split() or nltk.tokenize;
            in final version I'm using stanford-core-nlp pos_tagger to tokenize.

    The function mark all rows with an (currently) invalid prompt (i.e. wrong verb_form for the target noun).
    One should leverage the returned CSV to generate a new JSON for re-annotating the invalid prompts with
    the corrected ones. Use
    :param annot_df:
    :return: a DataFrame adding to annot_df:
        * an "invalid_prompt" column (boolean) - True if prompt changed (requires re-annotation or removal).
        * a "corrected_verb_form" column (string) indicating the up-to-date suggested verb_form returned by
        candidate_extraction.py. Notice that if it's an empty string, it means that for the current algorithm,
        no verb_form is available for this noun, meaning that we should delete the current annotation row but we should
        not re-annotate this target noun.
    """
    def annot_row_to_corrected_verb_form(row: pd.Series) -> str:
        from qanom.candidate_extraction import candidate_extraction
        noun = row.verb
        uptodate_verb_forms, is_had_uptodate_verbs = \
            candidate_extraction.get_verb_forms_from_lexical_resources(noun, **candidate_extraction.resources)
        # return empty string for non-verbal nouns (nouns with no verb_forms)
        if not is_had_uptodate_verbs:
            return ""
        else:
            # take the best verb_form
            return uptodate_verb_forms[0]

    # use annot_row_to_corrected_verb_form to generate the new columns
    annot_df["corrected_verb_form"] = annot_df.apply(annot_row_to_corrected_verb_form, axis=1)
    annot_df["invalid_prompt"] = annot_df.apply(lambda row: row["corrected_verb_form"] != row["verb_form"], axis=1)
    return annot_df


def reannotate_corrected_verb_forms(annot_df: pd.DataFrame, output_json_fn) -> NoReturn:
    """
    Generate input for the qasrl_crowdsourcing project (equivalent to output of candidate_extraction.py)
    :param annot_df: returned from find_invalid_prompts()
    :param output_json_fn: file name where to dump the JSON of the prompts for re-annotation
    """
    pred_idx_lbl = common.get_predicate_idx_label(annot_df)
    annot_df.drop_duplicates(subset=["qasrl_id", pred_idx_lbl])
    invalidated_df = annot_df[annot_df["invalid_prompt"]]    # filter only invalidated prompts
    re_annot_df = invalidated_df[invalidated_df["corrected_verb_form"]!=""] # filter out predicates now with no verb form
    candidates = [{"sentenceId": row["qasrl_id"],
                   "tokSent" : row["sentence"].split(" "),
                   "targetIdx": row[pred_idx_lbl],
                   "verbForms": [row["corrected_verb_form"]]}
                  for id, row in re_annot_df.iterrows()]
    json.dump(candidates, open(output_json_fn, "w"))


def replace_some_annotations(orig_df: pd.DataFrame, corrected_annot_df: pd.DataFrame) -> pd.DataFrame:
    """  Return orig_df after replacing the predicates that are in corrected_annot_df with their
         corrected annotation. """
    # replace intersection of predicates
    predicates_to_replace = set(corrected_annot_df.key.drop_duplicates()) & set(orig_df.key.drop_duplicates())
    # remove them from (a copy of) orig
    final_df = orig_df[~orig_df.key.isin(predicates_to_replace)].copy()
    # now merge with the annotations of these predicates in corrected_annot_df
    relevant_corrected_df = corrected_annot_df[corrected_annot_df.key.isin(predicates_to_replace)].copy()
    final_df = pd.concat([final_df, relevant_corrected_df], sort=False)
    return final_df


def fix_annot_with_corrected(orig_annot_fn: str, corrected_annot_fn: str,
                             dest_dir: str = "files/annotations/production/corrected") -> NoReturn:
    orig_df = read_annot_csv(orig_annot_fn)
    all_corrected_df = read_annot_csv(corrected_annot_fn)
    corrected_df = replace_some_annotations(orig_df, all_corrected_df)
    # in addition to re-annotation correction, filter out currently invalid prompts for data
    corrected_df = find_invalid_prompts(corrected_df)
    corrected_and_filtered_df = corrected_df[~corrected_df.invalid_prompt]
    final_df = corrected_and_filtered_df.drop(["corrected_verb_form", "invalid_prompt"], axis=1)
    # now export to file with same naming as orig (but in destination folder)
    orig_dir, orig_name = os.path.split(orig_annot_fn)
    dest_fn = os.path.join(dest_dir, orig_name)
    save_annot_csv(final_df, dest_fn)


def generate_corrected_annotation_files():
    """
    Fixing original annotations with new verb-form algorithm and the resulting output of the re-annotation batch.
    """
    orig_dir_path = "files/annotations/production"
    cor_path = "files/annotations/production/corrected"
    reannot_fn = "files/annotations/reannot.csv"
    ann_files = [os.path.join(orig_dir_path, fn) for fn in os.listdir(orig_dir_path) if fn.endswith(".csv")]
    for orig_fn in ann_files:
        fix_annot_with_corrected(orig_fn, reannot_fn, cor_path)

"""
Functions for filtering Non-Morpholigically-Related verb_forms (based on NMR feedbacks from workers)
Note: the methods that fix the annotations use the file files/nmr_case.csv.  This file 
 is a black-list containing all (noun, verb_form) pairs that should be removed, and it can be edited.
"""
nmr_cases_fn = 'files/nmr_cases.csv'


def generate_filtered_annotation_files():
    """
    Filtering annotations by removing NMR (noun, verb_form) cases.
    """
    # orig_dir_path = "files/annotations/gold_set/generation/corrected"
    # dest_path = "files/annotations/gold_set/generation/corrected_filtered"
    orig_dir_path = "files/annotations/train_set/orig"
    dest_path = "files/annotations/train_set/filtered"
    ann_files = [os.path.join(orig_dir_path, fn) for fn in os.listdir(orig_dir_path) if fn.endswith(".csv")]
    for orig_fn in ann_files:
        fix_annot_with_nmr_blacklist(orig_fn, dest_path)


def fix_annot_with_nmr_blacklist(orig_annot_fn: str,
                                 dest_dir: str) -> NoReturn:
    orig_df = read_annot_csv(orig_annot_fn)
    filtered_df = remove_NMR_cases_from_annotations(orig_df)
    # now export to file with same naming as orig (but in destination folder)
    orig_dir, orig_name = os.path.split(orig_annot_fn)
    dest_fn = os.path.join(dest_dir, orig_name)
    save_annot_csv(filtered_df, dest_fn)


def remove_NMR_cases_from_annotations(annot_df: pd.DataFrame) -> pd.DataFrame:
    """ Take all predicates with NMR feedbacks, and remove them. return the filtered df. """
    import csv
    with open(nmr_cases_fn) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        not_morphologically_related_cases = list(csv_reader)[1:]
    # incorporate both capital and non-capital to black-list of NMRs
    nmrs = set(utils.flatten([[(noun.capitalize(), verb),
                        (noun.casefold(), verb)]
                        for noun,verb in not_morphologically_related_cases]))
    # remove these from annot_df
    is_nmr = lambda r: (r.verb, r.verb_form) in nmrs
    is_nmr_series =  annot_df.apply(is_nmr, axis=1)
    print(f"filtering out {annot_df[is_nmr_series]['key'].drop_duplicates().shape[0]} Non-Morphologically-Related predicates ")
    filtered_df = annot_df[~is_nmr_series]
    return filtered_df


def set_sentence_column(df: pd.DataFrame, df_with_sentences: pd.DataFrame) -> NoReturn:
    """ Add 'sentence' to df by qasrl_id (inplace).
    Use df_with_sentences (assuming it has both qasrl_id and sentence columns) to take the sentence from. """
    sent_id2sent = dict(zip(df_with_sentences.qasrl_id, df_with_sentences.sentence))
    df['sentence'] = df.qasrl_id.apply(sent_id2sent.get)


def get_NMR_cases(feedbacks_dir: str) -> Set[Tuple[str, str]]:
    """ get all NMR cases - all (noun, verb_form) pairs that got any NMR feedback anywhere.
     The function would work for feedback files containing 'sentence' column.
     """
    fddf = read_dir_of_csv(feedbacks_dir, sep="\t")
    # fix feedback-files issues
    fix_qasrl_id(fddf)
    fddf['noun'] = fddf.apply(lambda r: r.sentence.split()[r.verb_idx], axis=1)
    annotations_with_nmr = fddf[fddf.feedback.astype(str).str.lower().str.contains("nmr")]
    nmr_cases = set(zip(annotations_with_nmr.noun, annotations_with_nmr.verb_form))
    return nmr_cases


def fix_qasrl_id(annot_df) -> NoReturn:
    """
    if the qasrl_id column is in the wrong form 'SentenceId(wikinews:99999:0:0)',
    fix it to 'wikinews:99999:0:0'.
    """
    annot_df['qasrl_id'] = annot_df.qasrl_id.apply(lambda s:s.lstrip('SentenceId(').rstrip(')'))


def set_feedback_column(annot_df: pd.DataFrame, feedbacks: pd.DataFrame) -> NoReturn:
    fix_qasrl_id(feedbacks)
    from qanom.annotations.common import set_key_column
    set_key_column(feedbacks)
    key2feedback = dict(zip(feedbacks.key, feedbacks.feedback))
    annot_df['feedback'] = annot_df.key.apply(key2feedback.get)


"""
Functions for adjusting the "duplicated" annotation to the complete dataset - select one worker triplet, 
and take their annotations as regular (non-duplicated) annotations.  
"""


def prune_duplicated_annot(gen_dupl_df: pd.DataFrame, arb_dupl_df: pd.DataFrame) -> pd.DataFrame:

    def get_annot_of_single_worker(df: pd.DataFrame) -> pd.DataFrame:
        """ For each predicate, get only annotations of one worker."""
        delim = '--'
        df['workers_str'] = df.groupby('key').worker_id.transform(lambda v: delim.join(set(v)))
        df['worker_index'] = df.apply(lambda r: r['workers_str'].split(delim).index(r.worker_id), axis=1)
        return df[df['worker_index'] == 0].drop(['workers_str', 'worker_index'], axis=1)

    sing_arb_df = get_annot_of_single_worker(arb_dupl_df)
    # now find those predicates that are in generation but not in single arbitration -
    # and take one "isVerbal==False" response into final csv
    keys_in_arb = set(sing_arb_df.key)
    gen_df_required = gen_dupl_df[(~gen_dupl_df.key.isin(keys_in_arb)) & (~gen_dupl_df.is_verbal)]
    # key only single row (response) per predicate
    gen_df_required = gen_df_required.drop_duplicates('key')
    combined_df = pd.concat([gen_df_required, sing_arb_df], ignore_index=True, sort=False)
    final_df = convert_to_final_annot(combined_df, get_anonymization(all_worker_ids))
    return final_df


def generate_pruned_dupl_annot() -> NoReturn:
    gen_dupl_fn = "files/annotations/gold_set/generation/corrected_filtered/annot.dupl.wikinews.dev.5.csv"
    arb_dupl_fn = "files/annotations/gold_set/arbitration/arbit.dupl.wikinews.dev.5.csv"
    out_fn = "files/annotations/gold_set/final/annot.final.wikinews.dev.5.csv"
    gen_dupl_df = read_annot_csv(gen_dupl_fn)
    arb_dupl_df = read_annot_csv(arb_dupl_fn)
    pruned_final_df = prune_duplicated_annot(gen_dupl_df, arb_dupl_df)
    save_annot_csv(pruned_final_df, out_fn)


"""
Functions for producing the final annotations - taking the .arbit file and adding the 
predicates with isVerbal==false,false from generation (that haven't been sent to consolidation). 
"""

# all workers participating in our annotation project
all_worker_ids = {'A21LONLNBOB8Q', 'A2A4UAFZ5LW71K',
                  'A3IR7DFEKLLLO', 'A98E8M4QLI9RS',
                  'AJQGWGESKQT4Y', 'A1FS8SBR4SDWYG',
                  'A25AX0DNHKJCQT'}


def generate_final_annotation_files() -> NoReturn:
    """
    Generating the final gold annotations -
    1. taking the .arbit file and adding the predicates
    with isVerbal==false,false from generation (that haven't been sent to consolidation).
    2. Anonymize worker-id
    3. Adjust CSV columns
    """
    gen_dir_path = "files/annotations/gold_set/generation/corrected_filtered"
    arb_dir_path = "files/annotations/gold_set/arbitration"
    dest_path = "files/annotations/gold_set/final"
    arb_name_to_gen_name = lambda name: '.'.join(['annot'] + name.split('.')[1:])
    ann_files = [(os.path.join(arb_dir_path, fn), os.path.join(gen_dir_path, arb_name_to_gen_name(fn)))
                 for fn in os.listdir(arb_dir_path)
                 if fn.endswith(".csv") and arb_name_to_gen_name(fn) in os.listdir(gen_dir_path)]
    # prepare worker anonymization (dataset-wide)
    anonymization: Dict[str, str] = get_anonymization(all_worker_ids)
    for arb_fn, gen_fn in ann_files:
        arb_df = read_annot_csv(arb_fn)
        gen_df = read_annot_csv(gen_fn)
        # combine arb with (false,false) predicates from gen
        combined_df = combine_to_final_annot(arb_df=arb_df, gen_df=gen_df)
        # make internal aesthetic modifications in the DataFrame
        final_df = convert_to_final_annot(combined_df, anonymization)
        # save
        fn = os.path.basename(arb_fn)
        # remove prefix and put new one
        fn = 'annot.final.' + fn.lstrip("arbit.")
        dest_fn = os.path.join(dest_path, fn)
        save_annot_csv(final_df, dest_fn)


def convert_to_final_annot(combined_df: pd.DataFrame, anonymization: Dict[str, str]) -> pd.DataFrame:
    """
    Process annot_df with final modifications for dataset publish -
        1. anonymize worker_id
        2. drop unnecessary columns - 'is_redundant', 'assign_id'
        3. rename columns - 'verb' -> 'noun';  'verb_idx' -> 'target_idx'
    @:argument combined_df: a complete annotation dataframe (that includes arbitrator annotation and the
    "is_verbal=False,False" predicates from generation.
    """
    # anonymize
    anon_df = anonymize_workers(combined_df, anonymization)
    # format data - drop unnecessary columns, rename
    columns_to_drop = {'is_redundant', 'assign_id', 'source_assign_id'} & set(anon_df.columns)
    final_df = anon_df.drop(columns=columns_to_drop)
    utils.rename_column(final_df, 'verb', 'noun')
    utils.rename_column(final_df, 'verb_idx', 'target_idx')
    return final_df


def combine_to_final_annot(arb_df: pd.DataFrame, gen_df: pd.DataFrame) -> pd.DataFrame:
    """
    Add the dropped predicates from generation (not passed to arb) to arbitration annotations.
    """
    # stay with onw row per predicate
    gen_df_shallow = gen_df.drop_duplicates(['key', 'worker_id'])
    # stay with those predicates with 2 generators
    key2ngen = dict(gen_df_shallow.key.value_counts())
    gen_df_shallow['ngen'] = gen_df_shallow.key.apply(key2ngen.get)
    gen_df_2gen = gen_df_shallow[gen_df_shallow.ngen==2]

    # don't fail on duplicated annotation file
    if gen_df_2gen.size==0:
        return arb_df
    # required predicates - those predicates which got no questions (by no generator)
    # these were not sent to arbitration - we have decided to declare them as "not verbal".
    gen_df_2gen = common.set_n_roles_per_predicate(gen_df_2gen)
    gen_df_required = gen_df_2gen[gen_df_2gen['n_roles'] == 0]
    # remove duplicates and drop auxiliary columns
    gen_df_required = gen_df_required.drop_duplicates('key').drop(['ngen', 'n_roles', 'no_roles'], axis=1)
    # combine gen with arb
    final_df = pd.concat([gen_df_required, arb_df], ignore_index=True, sort=False)
    return final_df


def get_anonymization(all_worker_ids: Set[str]) -> Dict[str, str]:
    import random
    random.seed(1000)
    wid_list : List[str] = list(sorted(all_worker_ids))
    random.shuffle(wid_list)
    wid2anon_wid = {wid : "Worker-"+str(wid_list.index(wid)+1)
                    for wid in wid_list}
    return wid2anon_wid


def anonymize_workers(annot_df: pd.DataFrame, worker_id_anonymization: Dict[str, str]) -> pd.DataFrame:
    res = annot_df.copy()
    workers_labels = [label for label in annot_df.columns if 'worker' in label]
    for lbl in workers_labels:
        for wid, new_wid in worker_id_anonymization.items():
            res[lbl] = res[lbl].astype(str).str.replace(wid, new_wid)
    return res


def generate_final_train_annotations() -> NoReturn:
    """
    Generating the final train-set annotations -
    1. Anonymize worker-id
    2. Adjust CSV columns
    """
    orig_train_dir_path = "files/annotations/train_set/filtered"
    dest_path = "files/annotations/train_set/final"
    ann_files = [os.path.join(orig_train_dir_path, fn)
                 for fn in os.listdir(orig_train_dir_path) if fn.endswith('.csv')]
    # prepare worker anonymization (dataset-wide)
    anonymization: Dict[str, str] = get_anonymization(all_worker_ids)
    for gen_fn in ann_files:
        gen_df = read_annot_csv(gen_fn)
        # make internal aesthetic modifications in the DataFrame
        final_df = convert_to_final_annot(gen_df, anonymization)
        # save
        fn = os.path.basename(gen_fn)
        dest_fn = os.path.join(dest_path, fn)
        save_annot_csv(final_df, dest_fn)


def merge_splits_into_final_large_files() -> NoReturn:
    gold_final_dir = "files/annotations/gold_set/final"
    train_final_dir = "files/annotations/train_set/final"

    wikinews_dev_fns = ["annot.final.wikinews.dev.1-4.csv",
                        "annot.final.wikinews.dev.5.csv"]
    wikinews_test_fns = ["annot.final.wikinews.test.1.csv",
                         "annot.final.wikinews.test.2.csv",
                         "annot.final.wikinews.test.3-5.csv"]
    wikipedia_test_fns = ["annot.final.wikipedia.test.1-5.csv"]
    wikipedia_dev_fns = ["annot.final.wikipedia.dev.1.1.csv",
                         "annot.final.wikipedia.dev.1.2-4.csv",
                         "annot.final.wikipedia.dev.2.csv",
                         "annot.final.wikipedia.dev.3-5.csv"]
    # gold data
    dir = gold_final_dir
    def concatCsvsInFinalDir(csv_fn_list: List[str], output_fn: str) -> NoReturn:
        csv_full_path_list = [os.path.join(dir, fn) for fn in csv_fn_list]
        output_path = os.path.join(dir, output_fn)
        utils.concatCsvs(csv_full_path_list, output_path)

    concatCsvsInFinalDir(wikinews_dev_fns, "annot.final.wikinews.dev.csv")
    concatCsvsInFinalDir(wikinews_test_fns, "annot.final.wikinews.test.csv")
    concatCsvsInFinalDir(wikipedia_dev_fns, "annot.final.wikipedia.dev.csv")
    concatCsvsInFinalDir(wikipedia_test_fns, "annot.final.wikipedia.test.csv")
    concatCsvsInFinalDir(["annot.final.wikinews.dev.csv", "annot.final.wikipedia.dev.csv"],
                         "annot.final.dev.csv")
    concatCsvsInFinalDir(["annot.final.wikinews.test.csv", "annot.final.wikipedia.test.csv"],
                         "annot.final.test.csv")

    # train data
    wikinews_train_fns = ["annot.wikinews.train.1.csv",
                          "annot.wikinews.train.2.csv",
                          "annot.wikinews.train.3.csv",
                          "annot.wikinews.train.4.csv",]
    wikipedia_train_fns = ["annot.wikipedia.train.1.csv",
                          "annot.wikipedia.train.2.csv",
                          "annot.wikipedia.train.3.csv",
                          "annot.wikipedia.train.4.csv",]
    dir = train_final_dir
    concatCsvsInFinalDir(wikinews_train_fns, "annot.wikinews.train.csv")
    concatCsvsInFinalDir(wikipedia_train_fns, "annot.wikipedia.train.csv")
    concatCsvsInFinalDir(["annot.wikinews.train.csv", "annot.wikipedia.train.csv"],
                         "annot.train.jsonl")


from typing import NoReturn, Dict, List, Iterator, Tuple

import numpy as np
import pandas as pd

from qanom import utils
from qanom.annotations.decode_encode_answers import Response, decode_response


def normalize(lst):
    a = np.array(lst)
    return a / sum(a)


def read_dir_of_csv(dir_path: str, prefix="", suffix="", sep=',') -> pd.DataFrame:
    """ Concatenate (all) csv files in directory into one DataFrame """
    import os
    dfs, sections = zip(*[(read_csv(os.path.join(dir_path, fn), sep=sep), fn.rstrip(".csv"))
                          for fn in os.listdir(dir_path) if fn.endswith(suffix+".csv") and fn.startswith(prefix)])
    return pd.concat(dfs, ignore_index=True, keys=sections, sort=False)


def read_dir_of_annot_csv(dir_path: str, prefix="", suffix="") -> pd.DataFrame:
    """ Concatenate (all) csv files in directory into one DataFrame """
    import os
    dfs, sections = zip(*[(read_annot_csv(os.path.join(dir_path, fn)), fn.rstrip(".csv"))
                          for fn in os.listdir(dir_path) if fn.endswith(suffix+".csv") and fn.startswith(prefix)])
    return pd.concat(dfs, ignore_index=True, keys=sections, sort=False)


def read_csv(file_path: str, sep=',') -> pd.DataFrame:
    try:
        df = pd.read_csv(file_path, sep=sep)
    except UnicodeDecodeError:
        df = pd.read_csv(file_path, encoding="Latin-1", sep=sep)
    return df


def read_annot_csv(file_path: str) -> pd.DataFrame:
    from qanom.annotations.decode_encode_answers import decode_qasrl
    df = read_csv(file_path)
    return decode_qasrl(df)


def save_annot_csv(annot_df: pd.DataFrame, file_path: str) -> NoReturn:
    from qanom.annotations.decode_encode_answers import encode_qasrl
    df = encode_qasrl(annot_df)
    df.to_csv(file_path, index=False, encoding="utf-8")


def iterate_qanom_responses(annot_df: pd.DataFrame) -> Iterator[Tuple[str, str, int, str, Response]]:
    """
    An iterator over every QANomResponse in the annotation-DataFrame.
    :param annot_df: 
    :return: Iterator, every instance is (sent_id, sent, target_index, worker_id, Response)
    """
    pred_idx_label = get_predicate_idx_label(annot_df)
    for key, predicate_df in annot_df.groupby('key'):
        for worker_id, assignment_df in predicate_df.groupby('worker_id'):
            response = decode_response(assignment_df)
            row = assignment_df.iloc[0]
            sentence_id, sentence, target_index = row[["qasrl_id", "sentence", pred_idx_label]]
            yield sentence_id, sentence, target_index, worker_id, response


""" Helper funcs for important information within an annotation DataFrame """


def get_predicate_idx_label(df: pd.DataFrame) -> str:
    """ Return name of the column of the predicate-index (changed in final annotation files to 'target_idx') """
    return (set(df.columns) & {'verb_idx', 'target_idx'}).pop()


def get_predicate_str_label(df: pd.DataFrame) -> str:
    """ Return name of the column of the predicate-string (changed in final annotation files to 'noun') """
    return (set(df.columns) & {'verb', 'noun'}).pop()


def set_key_column(annot_df: pd.DataFrame):
    """ Add 'key' column (predicate unique identifier) """
    pred_idx_label = get_predicate_idx_label(annot_df)
    if 'key' not in annot_df.columns and len(annot_df):
        annot_df['key'] = annot_df.apply(lambda r: r['qasrl_id']+"_"+str(r[pred_idx_label]), axis=1)


def set_sentence_columns(annot_df: pd.DataFrame, sentence_df: pd.DataFrame) -> NoReturn:
    """ Set a 'sentence' column to `annot_df` based on its `qasrl_id`. Retrieve sentence from `sentence_df`."""
    sent_map = get_sent_map(sentence_df)
    annot_df['sentence'] = annot_df.apply(lambda r: ' '.join(sent_map[r.qasrl_id]), axis=1)


def get_sent_map(annot_df: pd.DataFrame) -> Dict[str, List[str]]:
    sent_map = dict(zip(annot_df.qasrl_id, annot_df.sentence.apply(str.split)))
    return sent_map


def set_n_workers(df: pd.DataFrame) -> pd.DataFrame:
    # per predicate
    df['n_workers'] = df.groupby('key').worker_id.transform(pd.Series.nunique)
    return df


def set_n_roles(df: pd.DataFrame) -> pd.DataFrame:
    # per predicate per worker
    df['no_roles'] = df.groupby(['key', 'worker_id']).wh.transform(lambda x: x.isnull().sum())
    df['num_rows'] = df.groupby(['key', 'worker_id']).wh.transform(pd.Series.count)
    df['n_roles'] = df.apply(lambda r: r['num_rows']-r['no_roles'], axis=1)
    df = df.drop('num_rows', axis=1)
    return df


def set_n_roles_per_predicate(df: pd.DataFrame) -> pd.DataFrame:
    # per predicate, count roles joint from all workers
    df['no_roles'] = df.groupby('key').wh.transform(lambda x: utils.is_empty_string_series(x).sum())
    df['num_rows'] = df.groupby('key').wh.transform(pd.Series.count)
    df['n_roles'] = df.apply(lambda r: r['num_rows']-r['no_roles'], axis=1)
    df = df.drop(['num_rows'], axis=1)
    return df


def get_n_predicates(df: pd.DataFrame) -> int:
    # overall
    return df['key'].drop_duplicates().shape[0]


def get_n_assignments(df: pd.DataFrame) -> int:
    # get number of assignments captured within the CSV
    pred_idx_label = get_predicate_idx_label(df)
    return df[['qasrl_id', pred_idx_label, 'worker_id']].drop_duplicates().shape[0]


def get_n_positive_predicates(annot_df: pd.DataFrame) -> int:
    # gets a df of annotation, returns number of assignments where isVerbal==True.
    # for a single worker df, this is the number of positive predicates.
    # for a multi-worker annot-df, this should be divided by num of assignments to get the positive rate.
    reduced_df = annot_df.drop_duplicates(subset=["key", "worker_id"])
    n_positive_predicates = reduced_df.is_verbal.sum()
    return n_positive_predicates


def get_n_QAs(annot_df: pd.DataFrame) -> int:
    not_questions = utils.count_empty_string(annot_df.question)
    return annot_df.shape[0] - not_questions


def get_n_args(annot_df: pd.DataFrame) -> int:
    def count_non_empty(lst):
        return sum(1 for a in lst if a)
    return sum(annot_df.answer.apply(count_non_empty))


def filter_questions(annot_df: pd.DataFrame) -> pd.DataFrame:
    """ Return subset of `annot_df` with rows that corresponds to a non-empty question """
    with_q = annot_df[~utils.is_empty_string_series(annot_df.question)]
    return with_q


def filter_non_questions(annot_df: pd.DataFrame) -> pd.DataFrame:
    """ Return subset of `annot_df` with rows that corresponds to an empty question
    (which are either not-verbal or marked with "no-QA-applicable")"""
    from qanom import utils
    with_q = annot_df[utils.is_empty_string_series(annot_df.question)]
    return with_q


def get_n_argument_taking_predicates(annot_df: pd.DataFrame) -> int:
    with_q = filter_questions(annot_df)
    return get_n_positive_predicates(with_q)



from argparse import ArgumentParser
import sys, os

if __name__ == "__main__":
    ap = ArgumentParser()
    ap.add_argument("system_path", help="system predicted annotations (CSV file)")
    ap.add_argument("ground_truth_path", help="reference annotations (CSV file")
    ap.add_argument("sentences_path", default=None,
                    help="if the input files don't have a `sentence` column with the raw sentences (but only `qasrl_id`), provide a CSV file that maps `qasrl_id` to `sentence`.")
    args = ap.parse_args()

    from qanom.evaluation import evaluate
    evaluate.main(args.system_path, args.ground_truth_path, args.sentences_path)



import os
package_path = os.path.dirname(os.path.abspath(__file__)) # 'qanom' package base directory, e.g. "/home/.../site-packages/qanom" 
with open(os.path.join(package_path, "version.txt")) as f:
    __version__ = f.read().strip()


from typing import Iterable, Optional

from qanom.nominalization_detector import NominalizationDetector
from qanom.qasrl_seq2seq_pipeline import QASRL_Pipeline

qanom_models = {"baseline": "kleinay/qanom-seq2seq-model-baseline",
                "joint": "kleinay/qanom-seq2seq-model-joint"}  

default_detection_threshold = 0.7
default_model = "joint"

class QANomEndToEndPipeline():
    """
    This pipeline wraps the NominalizationDetector together with the Seq2Seq-based QANom parser.
    Given a sentence, the pipeline identifies nominal predicates (deverbal nominalizations), 
    and generates question-answer pairs capturing their predicate-argument relations.  
    """
    
    def __init__(self, 
                 qanom_model: Optional[str] = None, 
                 detection_threshold: Optional[float] = None):
        self.predicate_detector = NominalizationDetector()
        self.detection_threshold = detection_threshold or default_detection_threshold
        
        qanom_model = qanom_model or default_model
        model_url = qanom_models[qanom_model] if qanom_model in qanom_models else qanom_model
        self.qa_pipeline = QASRL_Pipeline(model_url)
    
    def __call__(self, sentences: Iterable[str], 
                 detection_threshold = None,
                 return_detection_probability = True,
                 **generate_kwargs):
        # get predicates
        threshold = detection_threshold or self.detection_threshold
        predicate_infos_for_sentences = self.predicate_detector(sentences, 
                                                                threshold=threshold,
                                                                return_probability=return_detection_probability)
        outputs = []
        for sentence, predicate_infos in zip(sentences, predicate_infos_for_sentences):
            # collect QAs for all predicates in sentence 
            predicates_full_infos = []
            for pred_info in predicate_infos:
                model_input = self._prepare_input_sentence(sentence, pred_info['predicate_idx'])
                model_output = self.qa_pipeline(model_input, 
                                                verb_form=pred_info['verb_form'], 
                                                predicate_type="nominal",
                                                **generate_kwargs)[0]
                predicates_full_info = dict(QAs=model_output['QAs'], **pred_info)
                predicates_full_infos.append(predicates_full_info)
            outputs.append(predicates_full_infos)
        return outputs
    
    def _prepare_input_sentence(self, raw_sentence: str, predicate_idx: int) -> str:
        words = raw_sentence.split(" ") 
        words = words[:predicate_idx] + ["<predicate>"] + words[predicate_idx:] 
        return " ".join(words)      
        

if __name__ == "__main__":
    pipe = QANomEndToEndPipeline(detection_threshold=0.75)
    sentence = "The construction of the officer 's building finished right after the beginning of the destruction of the previous construction ."
    print(pipe([sentence]))
    # Output:
    # [[{'QAs': [{'question': 'what was constructed ?',
    #      'answers': ["the officer 's"]}],
    #    'predicate_idx': 1,
    #    'predicate': 'construction',
    #    'predicate_detector_probability': 0.7623529434204102,
    #    'verb_form': 'construct'},
    #   {'QAs': [{'question': 'what began ?',
    #      'answers': ['the destruction of the']}],
    #    'predicate_idx': 11,
    #    'predicate': 'beginning',
    #    'predicate_detector_probability': 0.8923847675323486,
    #    'verb_form': 'begin'},
    #   {'QAs': [{'question': 'what was destructed ?', 
    #      'answers': ['the previous']}],
    #    'predicate_idx': 14,
    #    'predicate': 'destruction',
    #    'predicate_detector_probability': 0.849774956703186,
    #    'verb_form': 'destruct'}]]


import json
from collections import defaultdict
from typing import Dict, List, Iterable, Union, NoReturn, Tuple, Any

import pandas as pd


def almost_same(str1: str, str2: str) -> bool:
    from fuzzywuzzy import fuzz
    return fuzz.ratio(str1, str2)>90

def split_str_twice(string: str, del1: str, del2: str) -> Tuple[str, str, str]:
    """ Return a 3-element tuple by applying two splits on a str using two different delimiters.
        only regard the first occurrence of each delimiter.
      E.g.  split_str_twice("an_email@gmail.com", "_", "@") -> ("an", "email", "gmail.com")
      """
    first, second_and_third = string.split(del1,1)
    second, third = second_and_third.split(del2,1)
    return first, second, third

def is_empty_string_series(series: pd.Series) -> pd.Series:
    """ Return boolean Series capturing whether `series` is empty string or NA """
    series = series.fillna('')
    isEmpty = series==''
    return isEmpty

def count_empty_string(series: pd.Series) -> int:
    """ Return number of empty strings or NA in a Series """
    return is_empty_string_series(series).sum()

def rename_column(df: pd.DataFrame, orig_label: str, new_label: str, inplace=True):
    return df.rename(mapper={orig_label:new_label}, axis=1, inplace=inplace)

def concatCsvs(csv_fn_list: List[str], output_fn: str, columns: List[str] = None) -> NoReturn:
    inp_dfs = [pd.read_csv(fn) for fn in csv_fn_list]
    concatenated_df = pd.concat(inp_dfs, ignore_index=True, sort=False)
    # `columns` can determine subset (and order) of output columns
    if columns:
        concatenated_df = concatenated_df[columns]
    concatenated_df.to_csv(output_fn, index=False, encoding="utf-8")
    print(f"exported DataFrame with shape {concatenated_df.shape} to {output_fn}")
    return concatenated_df


def df_to_dict(df: pd.DataFrame, key_column: str, value_column: str) -> Dict[Any, Any]:
    """ Return a dict compiled from two columns in a DataFrame. """
    return dict(zip(df[key_column], df[value_column]))

def df_first_row(df: pd.DataFrame) -> pd.Series:
    return list(df.iterrows())[0][1]

def asRelative(distribution : Union[List, Dict]):
    # get a list\dict of numbers (a distribution), return the relative distribution (element/sum)
    if 'values' in dir(distribution):
        # a dict type
        sm = float(sum(distribution.values()))
        return {k: v / sm for k, v in distribution.items()}
    else:
        # a list type
        sm = float(sum(distribution))
        return [e / sm for e in distribution]

def replaceKeys(orig_dict, oldKeys2NewKeys, inplace=True):
    """ replace keys with new keys using oldKeys2NewKeys mapping. """
    target_dict = orig_dict if inplace else {}
    for oldKey, newKey in oldKeys2NewKeys.items():
        if oldKey in orig_dict:
            target_dict[newKey] = orig_dict.get(oldKey)
            if inplace: orig_dict.pop(oldKey)
    return target_dict

def removeKeys(orig_dict: Dict[Any, Any], keys: List[Any], inplace=True) -> Dict[Any, Any]:
    """ remove these keys from the orig_dict. """
    if not inplace:
        orig_dict = orig_dict.copy()
    for key in keys:
        orig_dict.pop(key)
    return orig_dict

def dictOfLists(pairs):
    # return a { key : [values given to that key] } for the pair list.
    # e.g. dictOfLists( [(0, "r"), (4, "s"), (0, "e")])  will return {0: ["r", "e"], 4: ["s"]}
    r = defaultdict(list)
    for k, v in pairs:
        r[k].append(v)
    return dict(r)


def majority(lst: Iterable[bool], whenTie=True) -> bool:
    lst = list(lst)
    s = sum(lst)
    if s == len(lst)/2.0:
        return whenTie
    else:
        return s > len(lst)/2.0

# jsonl (json-lines) util

class jsonl:
    @classmethod
    def dumps(cls, iterable: Iterable) -> str:
        return "\n".join(json.dumps(obj) for obj in iterable)

    @classmethod
    def dump(cls, iterable: Iterable, fp):
        s = jsonl.dumps(iterable)
        fp.write(s)

    @classmethod
    def loads(cls, lines_str: str) -> List[Any]:
        l = [json.loads(line) for line in lines_str.splitlines()]
        return l

    @classmethod
    def load(cls, fp) -> List[Any]:
        return jsonl.loads(fp.read())



# list, dict and sets utils
def is_iterable(e):
    return '__iter__' in dir(e)


def flatten(lst, recursively=False):
    """ Flatten a list.
    if recursively=True, flattens all levels of nesting, until reaching non-iterable items
    (strings are considered non-iterable to that matter.)
    :returns a flatten list (a non-nested list)
    """
    if not is_iterable(lst):
        return lst
    out = []
    for element in lst:
        if is_iterable(element):
            if recursively:
                out.extend(flatten(element))
            else:
                out.extend(element)
        else:
            out.append(element)
    return out


def is_nested(lst):
    return any(is_iterable(e) for e in lst)


def power_set(lst, as_list=True):
    """ set as_list to false in order to yield the power-set """
    import itertools
    pwset_chain = itertools.chain.from_iterable(itertools.combinations(lst, r)
                                                for r in range(len(lst) + 1))
    if as_list:
        return list(pwset_chain)
    else:
        return pwset_chain


def static_variables(**kwargs):
    """ A decorator for creating static local variables for a function.
    Usage Example:

    @static_variables(counter=0, large_list=load_large_list())
    def foo():
        foo.counter += 1    # now 'counter' and 'large_list' are properties of the method,
                            #  and are initialized only once, in the decorator line.
        print "Counter is %d" % foo.counter
        print foo.large_list[foo.counter]

    """
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate


def lazy_static_variables(**kwargs):
    """ A decorator for creating static local variables for a function with lazy initialization,
    i.e., variables are defined and initialized only
    Usage Example:

    @static_variables(counter=0, large_list=load_large_list())
    def foo():
        foo.counter += 1    # now 'counter' and 'large_list' are properties of the method,
                            #  and are initialized only once, in the decorator line.
        print "Counter is %d" % foo.counter
        print foo.large_list[foo.counter]

    """
    def decorate(func):
        print("decorator")
        for k in kwargs:
            setattr(func, k, kwargs[k]())
        return func
    return decorate



"""
A unified class for detecting "verbal" nominalizations from raw text -
wraps both `candidate_extraction` and `predicate_detector`
"""
from typing import Iterable, List, Dict, Any
from transformers import AutoModelForTokenClassification, AutoTokenizer
import pandas as pd

from qanom.candidate_extraction.candidate_extraction import extract_candidate_nouns

def dict_without(orig_dict: Dict[Any, Any], keys_to_remove: Iterable[Any]) -> Dict[Any, Any]:
    new_dict = orig_dict.copy()
    for key in keys_to_remove:
        if key in new_dict:
            new_dict.pop(key)
    return new_dict

class NominalizationDetector():    
    original_model_hub_name = "kleinay/nominalization-candidate-classifier"
    
    def __init__(self, model_hub_name = None):
        self.model_hub_name = model_hub_name or NominalizationDetector.original_model_hub_name
        self.model = AutoModelForTokenClassification.from_pretrained(self.model_hub_name)
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_hub_name)

    def __call__(self, raw_sentences: Iterable[str], 
                 return_all_candidates: bool = False, 
                 return_probability: bool = True,
                 threshold: float = 0.5):
        """
        Returns a list (sized as `raw_sentences`) of lists of nominalization infos (dicts). 
        """
        # use qanom's candidate_extraction module to identify candidates, i.e. nouns with morphologically-related verbs (using lexical resources)
        sentences_dict = {str(i): s for i,s in enumerate(raw_sentences)}
        candidate_infos: List[Dict[str, Any]] = extract_candidate_nouns(sentences_dict, "dict")
        # keys: 'sentenceId', 'tokSent', 'targetIdx', 'verbForms'
        
        # edge case - candidate_infos is totally empty
        if not candidate_infos:
            return [[]] * len(raw_sentences)
        cand_df = pd.DataFrame(candidate_infos)
        grouped_df = cand_df.groupby('sentenceId') 
        input_sent_df = pd.DataFrame()  # model is sentence level, so prepare input at sentence level
        input_sent_df["words"] = grouped_df.tokSent.first()
        input_sent_df["sent_idx"] = int(grouped_df.sentenceId.first())
        input_sent_df["target_idxs"] = grouped_df.targetIdx.apply(list)
        input_sent_df["verb_forms"] = grouped_df.verbForms.apply(lambda r: [l[0] for l in r])

        # since not all raw_sentences necessarily has candidates, must keep mapping to original sentence index
        included_sentences = [raw_sentences[i] for i in list(input_sent_df["sent_idx"])]
        included_idx2orig_idx = list(input_sent_df["sent_idx"])
        
        # just for reference:
        label_map = self.model.config.id2label
        positive_lbl_id = 0     # because label_map[0] == 'True' 
        negative_lbl_id = 1     # because label_map[1] == 'False' 
        
        tokenized = self.tokenizer(included_sentences, return_tensors="pt", padding=True)
        logits = self.model(tokenized.input_ids).logits
        # compose binary probability as the mean of positive label's prob with complement of neg. label's prob.
        positive_probability = logits[:,:,positive_lbl_id].sigmoid()   # probability of the "true" label (index 0 in axis-2)
        negative_probability = logits[:,:,negative_lbl_id].sigmoid()
        probability = (positive_probability + (1-negative_probability) ) / 2
        preds = probability > threshold

        batch_size, seq_len = preds.shape # same shape as probability
        assert batch_size == len(included_idx2orig_idx)
        
        predicate_lists = [[]] * len(raw_sentences) # final returned list is aligned with raw_sentences
        for i, orig_idx in enumerate(included_idx2orig_idx):
            # orig_idx is the sentence index in the input arg `raw_sentence`
            row = input_sent_df.iloc[i]
            # get mapping of word to tokens (for using indexes from dataset)
            wordId2numOfTokens = [len(self.tokenizer.tokenize(word)) for word in row.words]
            wordId2firstTokenId, curr_tok_id = [], 1
            for wordId in range(len(wordId2numOfTokens)):
                wordId2firstTokenId.append(curr_tok_id)
                curr_tok_id += wordId2numOfTokens[wordId]
            # get dict of all candidates to prediction (on first token of word)
            predicate_lists[orig_idx] = [
                {"predicate_idx": pred_idx,
                 "predicate": row.words[pred_idx],
                 "predicate_detector_prediction": preds[i][wordId2firstTokenId[pred_idx]].item(),
                 "predicate_detector_probability": probability[i][wordId2firstTokenId[pred_idx]].item(),
                 "verb_form": verb_form} 
                for pred_idx, verb_form in zip(row.target_idxs, row.verb_forms)]

        # filter nominalizations that were positively classified as carrying a verbal meaning in context
        if not return_all_candidates:
            keys_to_remove = ["predicate_detector_prediction"] + ([] if return_probability else ["predicate_detector_probability"])
            for i, sent_predicate_list in enumerate(predicate_lists):
                predicate_lists[i] = [dict_without(predicate_dict, keys_to_remove) 
                                      for predicate_dict in sent_predicate_list
                                      if predicate_dict["predicate_detector_prediction"]]
        return predicate_lists

if __name__ == "__main__":
    raw_sentences: Iterable[str] = ["the construction of the officer 's building finished right after the beginning of the destruction of the previous construction ."]
    detector = NominalizationDetector()
    from pprint import pprint
    pprint(detector(raw_sentences, return_all_candidates=True))
    pprint(detector(raw_sentences, threshold=0.75, return_probability=False))



"""
Use lexical resources to extract nouns that are candidates for being deverbal nominalizations.

The script can be used for generating a JSON file that serves as input to the qasrl-crowdsourcing MTurk application
for annotating verbal arguments of nominalizations (qanom-unified branch).
It can also be used for generating a CSV files that serves as input to the predicate_detector.


Logic building blocks:
* load raw data  (sentences) from corpus
* use Stanford CoreNLP to POS-tag the sentence
* get common nouns ("NN", "NNS")
* filter by wordnet + CatVar (+ possible-noms from the inhouse verb_to_nom utility)
* generate a list of nominalization candidates, each containing:
    { "sentenceId": sentence ID (string),
      "tokSent": sentence tokens (list),
      "targetIdx": index of candidate noun (int),
      "verbForms": non-empty list of verb-forms (strings) (ordered by edit distance)
    }

"""

import json, os, sys
from typing import List, Tuple, Iterable, Dict, Any

import pandas as pd
import nltk
from nltk.downloader import Downloader

# add project basic directory to sys.path, in order to refer qanom as a package from anywhere
project_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
sys.path.append(project_dir)

from qanom.candidate_extraction import cand_utils
from qanom.annotations.common import read_csv
from qanom.candidate_extraction.verb_to_nom import SuffixBasedNominalizationCandidates as VTN

""" Define which resources should be used (by default) for filtering nouns as candidate nominalizations. """
default_resources = {"wordnet": True,
                     "catvar": True,
                     "affixes_heuristic": True}

vtn = None # init this global VTN object only if required

# by default, use nltk's default pos_tagger ('averaged_perceptron_tagger'):
tagger_package = 'averaged_perceptron_tagger'
nltk_downloader = Downloader()
if (not nltk_downloader.is_installed(tagger_package)):
    nltk.download(tagger_package)
    
pos_tag = nltk.pos_tag
"""
Alternatively, when extracting candidates for crowdsourcing QANom annotations through the qasrl-crowdsourcing project,
one should use the same POS model as inside qasrl-crowdsourcing for consistency. 
qasrl-crowdsourcing uses the CoreNLPParser model in Java. We will use here nltk's CoreNLPParser wrapper.
To run the CoreNLPParser model as a server on your machine (port 9000), 
pre-run the following command from the unzipped directory of the stanford-core-nlp project 
(see https://www.khalidalnajjar.com/setup-use-stanford-corenlp-server-python/ for instructions): 
```bash
java -mx4g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -annotators "tokenize,ssplit,pos,lemma,parse,sentiment" -port 9000 -timeout 30000
```
Then, replace the above two-lines in the python script with the following block: 
```python
from nltk.parse import CoreNLPParser
pos_tagger = CoreNLPParser(url='http://localhost:9000', tagtype='pos')
pos_tag = pos_tagger.tag
```
"""

COMMON_NOUNS_POS = ["NN", "NNS"]

def levenshteinDistance(s1, s2):
    if len(s1) > len(s2):
        s1, s2 = s2, s1

    distances = range(len(s1) + 1)
    for i2, c2 in enumerate(s2):
        distances_ = [i2+1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]

# load data from corpus
def get_sentences_from_csv(csv_fn):
    """
    Load raw data (sentence with sentence-id) from simple csv format
    :param csv_fn: file name for the simple raw_data file (sentenceId, sentence)
    :return: dict {sentenceId : sentence}
    """
    df = read_csv(csv_fn)
    possible_labels_for_sent_id = {'sentenceId', 'sentence_id', 'qasrl_id'}
    sent_id_label = list(possible_labels_for_sent_id & set(df.columns))[0]
    df = df.set_index(sent_id_label)
    records = df[['sentence']].to_records()
    return dict(records)


def tokenize_and_pos_tag(sentence):
    """
    Apply CoreNLP POS-Tagger on sentences.
    CoreNLP is running on server, so use it to be compatible with the qasrl-crowdsourcing POS tagging
    :param sentence: unicode string
    :return:
     pos is [ (token, POS) for each token in sentence ]
     tok is [ tok for each token in sentence ]
     returning (tok, pos)
    """
    pos_tagged = list(pos_tag(sentence.split(" ")))
    tokenized = [e[0] for e in pos_tagged]
    return tokenized, pos_tagged


# get common nouns ("NN", "NNS")
def get_common_nouns(posTaggedSent: List[Tuple[str, str]]):
    """
    Filter and retrieve common-nouns from the sentence (to be candidates for nominaliation)
    :param posTaggedSent: [ (token1, pos1), ...]
    :return: list of (index, token) tuples
    """
    nouns = []
    for index, (token, pos) in enumerate(posTaggedSent):
        if pos in COMMON_NOUNS_POS:
            nouns.append((index, token))
    return nouns


# verbalize using both WordNet, CatVar and our in-house suffix-based
# heuristic (=VTN: Verb To possible Nominalizations; implementation at `verb_to_nom.py`)
def get_verb_forms_from_lexical_resources(nn,
                                          wordnet=True,
                                          catvar=True,
                                          affixes_heuristic=True,
                                          filter_distant=False) -> Tuple[List[str], bool]:

    wordnet_verbs = []
    if wordnet:
        from qanom.candidate_extraction import wordnet_util
        wordnet_verbs = wordnet_util.convert_pos_by_lemmas(nn, wordnet_util.WN_NOUN, wordnet_util.WN_VERB)
        # apply distant-verbs filtering on wordnet verbs
        if filter_distant:
            wordnet_verbs = list(filter(lambda v: filter_distant_verb_forms(v, nn), wordnet_verbs))

    catvar_verbs = []
    if catvar:
        from qanom.candidate_extraction import catvar as catvar_util
        catvar_verbs = catvar_util.catvariate(nn)

    affixes_heuristic_verbs = []
    if affixes_heuristic:
        global vtn
        if vtn is None: # initialize global vtn only once, if required
            vtn = VTN()
        affixes_heuristic_verbs = vtn.get_source_verbs(nn)
    # sort by distance
    vrbs = catvar_verbs + wordnet_verbs + affixes_heuristic_verbs
    vrbs = [v for v, w in cand_utils.results_by_edit_distance(nn, vrbs)]
    if vrbs:
        return vrbs, True
    else:
        return [nn], False


def is_candidate_noun(nn: str, **resources) -> bool:
    # does it have candidate (morphologically related) verb-forms?
    return get_verb_forms_from_lexical_resources(nn, **resources)[1]


def filter_distant_verb_forms(verb_form, noun):
    """ Return False for a verb_form whose edit-distance from noun is too large
    (indicating it's probably not morphologically related). """
    edit_distance = cand_utils.levenshteinDistance(verb_form, noun)
    short_size, long_size = min(len(verb_form),len(noun)), max(len(verb_form),len(noun))
    edit_distance_without_suffix = cand_utils.levenshteinDistance(verb_form[:short_size], noun[:short_size])
    avg_size = (len(verb_form)+len(noun))/2.0
    num_chars_maybe_arbitrarily_identical = short_size/3
    if edit_distance_without_suffix > short_size - num_chars_maybe_arbitrarily_identical:
        print("filtered out: +'"+verb_form+"' (edit-distance: "+str(edit_distance)+") for the noun '"+noun+"'")
        return False
    else:
        return True

"""
Different functions for different input formats of sentences.
"""
def get_sentences_from_allennlp_jsonl(sentences_josnl_fn: str) -> Dict[str, str]:
    """
    @:param sentences_josnl_fn: a file name of JSON-lines of input sentences, as allennlp predictors input format:
        {"sentence": <actual sentence string> }
        {"sentence": <actual sentence string> }
        ...
    Returns: dict of sentences. Generated a unique sentence_id for the sentences (by index).
    """
    with open(sentences_josnl_fn) as f:
        sentences = []
        for line in f.readlines():
            sentences.append(json.loads(line)["sentence"])
    return get_sentences_from_iterable(sentences)


def get_sentences_from_iterable(sentences: Iterable[str]) -> Dict[str, str]:
    """
    @:param sentences: iterable of raw sentences
    Returns: dict of sentences. Generated a unique sentence_id for the sentences (by index).
    """
    sentences_dict: Dict[str, str] = {f"Sentence-{1+i:04}": sent for i, sent in enumerate(sentences)}
    return sentences_dict


# the "core" function of extracting candidates from sentences
def get_candidate_nouns(sentences: Dict[str, str], **resources) -> List[Dict[str, Any]]:
    """
    @:param sentences: {sentence_id : sentence_string}
    Returns: list of candidate_info (dict)
    """
    all_candidates = []
    for sid, sentence in sentences.items():
        # POS-tagging and tokenize are dependant, so doing both together
        tokenizedSent, posTaggedSent = tokenize_and_pos_tag(sentence)
        for idx, nn in get_common_nouns(posTaggedSent):
            verb_forms, is_had_verbs = get_verb_forms_from_lexical_resources(nn, **resources)
            # take only common nouns that have optional verb-forms as candidates:
            if is_had_verbs:
                candidate_info = {"sentenceId": sid,
                                  "tokSent" : tokenizedSent,
                                  "targetIdx": idx,
                                  "verbForms": verb_forms[:5]}

                all_candidates.append(candidate_info)
    return all_candidates


def export_candidate_info_to_csv(candidates_info: List[Dict[str, Any]], csv_out_fn: str = None) -> pd.DataFrame:
    """
    Export the noun-candidates info (returned by `get_candidate_nouns`) into a QANom-csv format,
     with columns: 'qasrl_id', 'sentence', 'target_idx', 'noun'.
    It is useful when you want to use a candidate extraction heuristic as a pre-processing for QANom model
    that was trained on QANom data (which is in qanom csv format).

    :param candidates_info: see output of `get_candidate_nouns`
    :param csv_out_fn: [optional] a file name to write the csv into.
    :return: a DataFrame of the produced CSV
    """
    df = pd.DataFrame(candidates_info)
    # now modify the dataFrame to match qanom annotation format
    df['sentence'] = df.apply(lambda r: ' '.join(r['tokSent']), axis=1)
    df['sentence'] = df.apply(lambda r: ' '.join(r['tokSent']), axis=1)
    df['noun'] = df.apply(lambda r: r['tokSent'][int(r['targetIdx'])], axis=1)
    df['verb_form'] = df.apply(lambda r: r['verbForms'][0], axis=1)
    df = df.drop(['tokSent', 'verbForms'], axis='columns')
    df = df.rename(mapper={'targetIdx': 'target_idx',
                           'sentenceId': 'qasrl_id'},
                   axis='columns')
    if csv_out_fn:
        df.to_csv(csv_out_fn, index=False)
    return df

# a Utility function to use candidate extraction both as a module as well as script
def extract_candidate_nouns(input: Any,
                            input_format: str,  # Literal["iterable", "raw", "dict", "csv", "jsonl"]
                            **resources) -> List[Dict[str, Any]]:
    # get sentences-info into a {sentence_id: sentence} dict
    if input_format == "dict":
        sentences = input
    elif input_format == "iterable":
        sentences = get_sentences_from_iterable(input)
    elif input_format == 'raw':
        with open(input) as fin:
            sentences = get_sentences_from_iterable(fin.read().splitlines())
    elif input_format == 'csv':
        sentences = get_sentences_from_csv(input)
    elif input_format == 'jsonl':
        sentences = get_sentences_from_allennlp_jsonl(input)
    else:
        raise NotImplementedError('input_format must be "iterable", "dict", "csv" or "jsonl".')

    candidates = get_candidate_nouns(sentences, **resources)
    return candidates


def main(args):
    """ Expecting command line arguments from `argparse`. """

    # determine resources
    resources = {"wordnet": args.wordnet,
                 "catvar": args.catvar,
                 "affixes_heuristic": args.affixes_heuristic}

    candidates = extract_candidate_nouns(args.sentences_fn,
                                         args.input_format,
                                         **resources)

    # Write candidates information to output file. """
    if args.output_format == 'json':
        json.dump(candidates, open(args.output_fn, "w"))
    elif args.output_format == 'csv':
        export_candidate_info_to_csv(candidates, args.output_fn)




import os, sys
_current_module = sys.modules[__name__]
qanom_package_root_path = os.path.dirname(os.path.abspath(_current_module.__file__))
resources_path = os.path.join(qanom_package_root_path, 'resources')


"""
For clarification:  we use the terms "encoding" and "decoding" for the bidirectional path between two formats
of the qanom data. This script and the functions within it refer to the *exported CSV file* as the "encoded" format,
and to the code-internal annot_df format (in which some columns values are objects\instances, and not strings)
as the "decoded" format. The decoded format is easily transferrable between DataFrame and the QANom class hierarchy
defined in this file (Response, Role and Question. Argument is actually just an alias for Tuple[int, int]).
Main differences:
* a single answer span:
 ** encoded: answer-spans are strings in the form f"{start}:{end}" , e.g. "2:3"
 ** decoded: answer-spans are objects of type Tuple[int, int] , e.g. (2,3)
* answer / answer_range columns:
 ** encoded: "he~!~John", "2:3~!~7:8"
 ** decoded: ["he", "John"], [(2,3), (7,8)]
"""

from dataclasses import dataclass, field, asdict
from typing import List, Iterable, Tuple, Generator

import pandas as pd

SPAN_SEPARATOR = "~!~"
NO_RANGE = ""
INVALID = "INVALID"

Argument = Tuple[int, int]
def arg_length(argument: Argument) -> int:
    return argument[1]-argument[0]

QUESTION_FIELDS = ['wh', 'subj', 'obj', 'aux', 'verb_prefix', 'verb_slot_inflection', 'prep', 'obj2', 'is_passive', 'is_negated']


@dataclass(frozen=True)
class Question:
    text: str
    wh: str
    subj: str
    obj: str
    aux: str
    verb_prefix: str
    verb_slot_inflection: str
    prep: str
    obj2: str
    is_passive: bool
    is_negated: bool

    def __str__(self):
        return self.text

    def isEmpty(self) -> bool:
        return self.text == ""

    @classmethod
    def empty(cls):
        return Question("", "", "", "", "", "", "", "", "", False, False)

    @classmethod
    def text_only(cls, question_text):
        return Question(question_text, "", "", "", "", "", "", "", "", False, False)


@dataclass(frozen=True)
class Role:
    question: Question
    arguments: Tuple[Argument, ...]

    def text(self):
        return self.question.text


@dataclass(frozen=True)
class Response:
    is_verbal: bool
    verb_form: str
    roles: List[Role] = field(default_factory=list)

    def all_args(self) -> List[Argument]:
        return [arg for role in self.roles for arg in role.arguments]


def is_invalid_range(r: Argument):
    return r[0] == NO_RANGE


def encode_argument(arg: Iterable[Argument]) -> str:
    if not arg or list(arg)[0] == NO_RANGE:
        return NO_RANGE

    return SPAN_SEPARATOR.join([encode_span(span) for span in arg])


def encode_span(span: Argument):
    if span == NO_RANGE or not span:
        return NO_RANGE
    return "{}:{}".format(span[0], span[1])


def encode_argument_text(arg_text: Iterable[str]):
    return SPAN_SEPARATOR.join(arg_text)


def encode_argument_text_from_spans(arguments: Iterable[Argument], tokens: List[str]) -> str:
    """ Return a string for the CSV field 'answer' """
    return SPAN_SEPARATOR.join([span_to_text(span, tokens) for span in arguments])


def span_to_text(span: Argument, tokens: List[str]) -> str:
    if type(span) is not tuple:
        return ""
    span_start, span_end = span
    return " ".join(tokens[span_start: span_end])


def decode_span(span_str: str) -> Argument:
    splits = span_str.split(":")

    return int(splits[0]), int(splits[1])


def decode_argument(arg_str: str) -> List[Argument]:
    if not arg_str:
        return []
    ranges = arg_str.split(SPAN_SEPARATOR)
    if ranges[0] == NO_RANGE:
        return ranges

    ranges = [decode_span(span_str) for span_str in ranges]
    return ranges


def decode_qasrl(qasrl_df: pd.DataFrame) -> pd.DataFrame:
    # WHY WHY WHY WE HAVE NULLS??? (see below why)
    # qasrl_df.dropna(subset=['qasrl_id', 'target_idx', 'question'], inplace=True)
    cols = set(qasrl_df.columns)
    answer_range_cols = set([col for col in cols if "answer_range" in col])
    answer_cols = set([col for col in cols if "answer" in col]) - answer_range_cols
    # We sometimes have null values in answer (but not in answer range) if the answer was actually the string:
    # NA or anything in the default NA values for read_csv
    ## common NA values
    # no longer excluding inf representations
    # '1.#INF','-1.#INF', '1.#INF000000',
    # _NA_VALUES = set([
    # '-1.#IND', '1.#QNAN', '1.#IND', '-1.#QNAN', '#N/A N/A', '#N/A',
    # 'N/A', 'n/a', 'NA', '#NA', 'NULL', 'null', 'NaN', '-NaN', 'nan', '-nan', ''
    # ])
    # drop them, for now

    #if answer_cols:
    #    qasrl_df.dropna(subset=answer_cols, inplace=True)

    for c in answer_cols:
        # qasrl_df[c] = qasrl_df[c].astype(str)
        qasrl_df[c].fillna("", inplace=True)
        qasrl_df[c] = qasrl_df[c].apply(lambda a: a.split(SPAN_SEPARATOR))
    for c in answer_range_cols:
        # qasrl_df[c] = qasrl_df[c].astype(str)
        qasrl_df[c].fillna("", inplace=True)
        qasrl_df[c] = qasrl_df[c].apply(decode_argument)

    for c in QUESTION_FIELDS:
        if c in qasrl_df:
            qasrl_df[c].fillna("", inplace=True)

    from qanom.annotations.common import set_key_column
    set_key_column(qasrl_df)
    return qasrl_df


def encode_qasrl(qasrl_df) -> pd.DataFrame:
    for_csv = qasrl_df.copy()
    cols = set(qasrl_df.columns)
    answer_range_cols = set([col for col in cols if "answer_range" in col])
    answer_cols = set([col for col in cols  if "answer" in col]) - answer_range_cols

    for c in answer_range_cols:
        for_csv[c] = for_csv[c].apply(encode_argument)
    for c in answer_cols:
        for_csv[c] = for_csv[c].apply(encode_argument_text)
    return for_csv


def encode_response(response: Response, sentence_str: str, **kwargs) -> pd.DataFrame:
    """
    should return a dataFrame capturing the annotations in Response.
    of course, this df will only include annotation-columns and not (other) data- or metadata- columns.
    kwargs can be used to insert other information to the columns of the output dataframe - these key:value
    pairs will be part of the information that is the same in all the rows of this response (e.g. target_idx).
    """
    common_for_rows = dict({'is_verbal' : response.is_verbal, 'verb_form': response.verb_form}, **kwargs)
    def role2rowDict(role: Role) -> dict:
        questionDict = asdict(role.question)
        questionDict['question'] = questionDict.pop('text')
        qaDict = dict(questionDict,
                      answer_range=encode_argument(role.arguments),
                      answer=encode_argument_text_from_spans(role.arguments, sentence_str.split(" ")))
        rowDict = dict(qaDict, **common_for_rows)
        return rowDict

    rolesToEncode = response.roles if response.roles else [Role(Question.empty(), tuple())]
    lstOfRowDicts = list(map(role2rowDict, rolesToEncode))
    return pd.DataFrame(lstOfRowDicts)


def decode_response(predicate_df: pd.DataFrame) -> Response:
    roles = list(yield_roles(predicate_df))
    is_verbal = predicate_df['is_verbal'].iloc[0] if len(predicate_df)>0 else True
    verb_form = predicate_df['verb_form'].iloc[0] if len(predicate_df)>0 else ""
    return Response(is_verbal,
                    verb_form,
                    roles)


def question_from_row(row: pd.Series) -> Question:
    question_as_dict = {question_field: row[question_field]
                        for question_field in QUESTION_FIELDS}
    question_as_dict['text'] = row.question if not pd.isnull(row.question) else ""
    return Question(**question_as_dict)


def yield_roles(predicate_df: pd.DataFrame) -> Generator[Role, None, None]:
    for row_idx, role_row in predicate_df.iterrows():
        if any(q_field not in predicate_df.columns for q_field in QUESTION_FIELDS):
            # this is not a qanom annotation file, but rather other format (e.g. NomBank csv file).
            # Thus, take only question text as Question
            question = Question.text_only(role_row.question)
        else:
            question = question_from_row(role_row)
        arguments: List[Argument] = role_row.answer_range
        # for No-QA-Applicable, yield no role (empty list of roles)
        if not question.isEmpty():
            yield Role(question, tuple(arguments))


# (C) Gabriel Stanovsky
from nltk.corpus import wordnet as wn
from qanom.candidate_extraction import cand_utils
# Just to make it a bit more readable
WN_NOUN = 'n'
WN_VERB = 'v'
WN_ADJECTIVE = 'a'
WN_ADJECTIVE_SATELLITE = 's'
WN_ADVERB = 'r'


def convert_pos_by_lemmas(word, from_pos, to_pos):
    """ Transform words given from/to POS tags """

    lemmas = wn.lemmas(word, pos=from_pos)
    # Word not found
    if not lemmas:
        return []

    # Get related forms
    derivationally_related_forms = [(l, l.derivationally_related_forms()) for l in lemmas]

    # filter only the desired pos (consider 'a' and 's' equivalent)
    related_noun_lemmas = []

    for drf in derivationally_related_forms:
        for l in drf[1]:
            if l.synset().name().split('.')[1] == to_pos or \
                (to_pos in (WN_ADJECTIVE, WN_ADJECTIVE_SATELLITE)
                 and l.synset().name().split('.')[1] in (WN_ADJECTIVE, WN_ADJECTIVE_SATELLITE)):
                related_noun_lemmas += [l]

    # Extract the words from the lemmas
    related_words = [l.name() for l in related_noun_lemmas]
    return list(set(related_words))


def convert_pos_by_synset(word, from_pos, to_pos):
    """ Transform words given from/to POS tags """

    synsets = wn.synsets(word, pos=from_pos)
    # Word not found
    if not synsets:
        return []

    # Get all lemmas of the word (consider 'a'and 's' equivalent)
    lemmas = []
    for s in synsets:
        for l in s.lemmas():
            if s.name().split('.')[1] == from_pos or from_pos in (WN_ADJECTIVE, WN_ADJECTIVE_SATELLITE) and s.name().split('.')[1] in (WN_ADJECTIVE, WN_ADJECTIVE_SATELLITE):
                lemmas += [l]

    # Get related forms
    derivationally_related_forms = [(l, l.derivationally_related_forms()) for l in lemmas]

    # filter only the desired pos (consider 'a' and 's' equivalent)
    related_noun_lemmas = []

    for drf in derivationally_related_forms:
        for l in drf[1]:
            if l.synset().name().split('.')[1] == to_pos or \
                (to_pos in (WN_ADJECTIVE, WN_ADJECTIVE_SATELLITE)
                 and l.synset().name().split('.')[1] in (WN_ADJECTIVE, WN_ADJECTIVE_SATELLITE)):
                related_noun_lemmas += [l]

    # Extract the words from the lemmas
    related_words = [l.name() for l in related_noun_lemmas]
    return list(set(related_words))




def verbalize(word, from_pos=WN_NOUN):
    related_verbs = convert_pos_by_lemmas(word, from_pos, WN_VERB)
    return cand_utils.results_by_edit_distance(word, related_verbs)


if __name__ == "__main__":
    # example usage
    print(convert_pos_by_lemmas("recognition", WN_NOUN, WN_VERB))



"""
Utils for converting between QANom CSV format (and similarly QASRL-GS format as in (Roit et. al., 2020))
and Large-Scale-QA-SRL jsonl format (https://github.com/uwnlp/qasrl-bank/blob/master/FORMAT.md).
"""
import json, os
from dataclasses import asdict
from typing import Dict, Any, NoReturn, List, Tuple

import numpy as np
import pandas as pd

JSON_OBJ = Dict[str, Any]

from qanom.annotations.common import save_annot_csv, read_annot_csv, get_predicate_idx_label
from qanom.annotations.decode_encode_answers import SPAN_SEPARATOR, decode_qasrl, Question, Argument, question_from_row
from qanom.utils import jsonl, df_first_row, replaceKeys, removeKeys


"""
Export jsonl format to csv format 
"""
def jsonl_file_to_csv(qasrl_v2_fn: str, dest_dir: str) -> NoReturn:
    with open(qasrl_v2_fn, encoding='latin-1') as f:
        annot_df = pd.concat([sentence_json_to_df(json.loads(jline))
                              for jline in f],
                             ignore_index=True, sort=False)
    # save df in destination
    orig_dir, orig_name = os.path.split(qasrl_v2_fn)
    new_name = '.'.join(orig_name.split('.')[:-1]) + ".csv"
    dest_fn = os.path.join(dest_dir, new_name)
    save_annot_csv(decode_qasrl(annot_df), dest_fn)


def sentence_json_to_df(qasrlv2: JSON_OBJ) -> pd.DataFrame:
    sent_id = qasrlv2['sentenceId']
    tok_sent = qasrlv2['sentenceTokens']
    sentence = ' '.join(tok_sent)

    from qanom.annotations.decode_encode_answers import Response, Role, Question, Argument, encode_response
    data_columns = ['qasrl_id', 'sentence', 'verb_idx', 'key', 'verb', 'verb_form']
    from qanom.annotations.consolidation import FINAL_COLUMNS

    def verb_entry_to_df(vEntry: JSON_OBJ) -> pd.DataFrame:
        verbIdx = vEntry['verbIndex']
        verb = tok_sent[verbIdx]
        qaJsons = list(vEntry['questionLabels'].values())
        generator_id = qaJsons[0]['questionSources'][0] if qaJsons else ""

        # construct Response. then use it
        def qEntry_to_role(qEntry: JSON_OBJ) -> Role:
            # first get Question
            qDict : JSON_OBJ = qEntry['questionSlots']
            replce_underscore = lambda s: "" if s=="_" else s
            qDict = {k:replce_underscore(v) for k,v in qDict.items()}
            qDict["verb_slot_inflection"] = qDict.pop('verb')
            qDict.update(is_negated=qEntry['isNegated'],
                         is_passive=qEntry['isPassive'],
                         text=qEntry['questionString'],
                         verb_prefix="")
            question = Question(**qDict)
            # get answers - insert all answer spans from all validators even though it is overlapping and redundant)
            answer_spans : List[Argument] = [tuple(span)
                                             for ansJdg in qEntry['answerJudgments'] if 'spans' in ansJdg
                                             for span in ansJdg['spans']]
            # answer spans in JSON format are Tuple[int, int], same as defined by us as Argument
            return Role(question=question, arguments=tuple(answer_spans))

        qas: List[Role] = [qEntry_to_role(qa) for qa in qaJsons]
        response = Response(is_verbal=True, verb_form=verb, roles=qas)
        pred_df = encode_response(response, sentence)
        other_data = {'qasrl_id': sent_id,
                      'sentence': sentence,
                      'verb_idx': verbIdx,
                      'key': f"{sent_id}_{verbIdx}",
                      'verb': verb,
                      'verb_form': "",
                      'worker_id': generator_id,
                      'assign_id': ""
                      }
        pred_df = pred_df.assign(**other_data)
        return pred_df

    # concat all predicates of sentence
    verb_dfs = [verb_entry_to_df(entry) for entry in qasrlv2['verbEntries'].values()]
    if verb_dfs:
        df = pd.concat(verb_dfs,
                       ignore_index=True, sort=False)
    else:
        df = pd.DataFrame(columns=FINAL_COLUMNS)
    return df[FINAL_COLUMNS]


"""
Export csv format to jsonl format
"""
def qanom_csv_file_to_jsonl(qanom_csv_fn: str, dest_dir: str) -> NoReturn:
    annot_df = read_annot_csv(qanom_csv_fn)
    sentences_dicts = (sentence_df_to_sentence_jsonl_dict(sentence_df)
                       for qasrl_id, sentence_df in annot_df.groupby('qasrl_id'))
    # save jsonl in destination
    orig_dir, orig_name = os.path.split(qanom_csv_fn)
    new_name = '.'.join(orig_name.split('.')[:-1]) + ".jsonl"
    dest_fn = os.path.join(dest_dir, new_name)
    jsonl.dump(sentences_dicts, open(dest_fn, "w"))


def sentence_df_to_sentence_jsonl_dict(sent_df: pd.DataFrame) -> JSON_OBJ:
    obj = {}
    row: pd.Series = df_first_row(sent_df)
    obj['sentenceId'] = row['qasrl_id']
    obj['sentenceTokens'] = row['sentence'].split(" ")
    obj['verbEntries'] = dict(predicate_df_to_verb_entry(predicate_df)
                              for _, predicate_df in sent_df.groupby('key'))
    return obj


def predicate_df_to_verb_entry(predicate_df: pd.DataFrame) -> Tuple[str, JSON_OBJ]:
    """ Returns (predicate_index, predicate_entry), where predicate_entry is a dict
        with the keys: 'verbIndex', 'verbInflectedForms', 'questionLabels'. """
    obj = {}
    row: pd.Series = df_first_row(predicate_df)
    pred_idx_lbl = get_predicate_idx_label(predicate_df)
    obj['verbIndex'] = int(row[pred_idx_lbl])
    # additional information, only relevant for QANom and not QA-SRL:
    #   'isVerbal' - whether the candidate predicate is indeed a verbal predicate
    #       (otherwise, no QAs are annotated).
    #   'verbForm' - what is the verbal form (infinitive) of the nominalization
    obj['isVerbal'] = row.is_verbal
    obj['verbForm'] = row.verb_form
    if row.is_verbal:
        obj['verbInflectedForms'] = {} # todo, based on obj['verbForm']
        question_entries = [question_row_to_question_label(r)
                            for _, r in predicate_df.iterrows()]
        obj['questionLabels'] = {question_entry["questionString"]: question_entry
                                 for question_entry in question_entries}
    else:
        # negative instances (non verbal) would have empty questionLabels dict
        obj['verbInflectedForms'] = {}
        obj['questionLabels'] = {}
    return str(row[pred_idx_lbl]), obj


def question_row_to_question_label(row: pd.Series) -> JSON_OBJ:
    obj = {'questionString': row.question,
           'questionSources': [row.worker_id]
                                if 'source_worker_ids' not in row or np.isnan(row['source_worker_ids'])
                                else row['source_worker_ids'].split(SPAN_SEPARATOR),
           'isNegated': row.is_negated,
           'isPassive': row.is_passive,
           }
    # qasrl-v2 also have 'tense', 'isPerfect' and 'isProgressive' keys.
    # but currently it doesn't seem important to recover them from the qanom data.
    # if we would like to do that, the information is probably at 'verb_slot_inflection' column.
    question: Question = question_from_row(row)
    questionDict = asdict(question)
    replaceKeys(questionDict, {"verb_slot_inflection": "verb"})
    removeKeys(questionDict, ["text", "is_passive", "is_negated"])
    obj['questionSlots'] = questionDict
    # answers
    answer_spans: List[Argument] = row.answer_range
    # qanom or qasrl-gs data would only have a single answerJudgment because data is consolidated
    obj['answerJudgments'] = [{'sourceId': row.worker_id,
                               'isValid': True,
                               "spans": answer_spans}]
    return obj



import os
from argparse import ArgumentParser

import pandas as pd

from qanom.annotations.common import read_annot_csv, read_csv
from qanom.evaluation.evaluate import eval_datasets


def parse_args():
    ap = ArgumentParser()
    ap.add_argument("qasrl_path", help="/path/to/qasrl_annotation_output.csv")
    ap.add_argument("ref_path", help="/path/to/qasrl_ground_truth.csv")
    ap.add_argument("sent_path", help="/path/to/sentences.csv")
    ap.add_argument("out_dir", help="/path/to/directory_where_a_report_for_each_worker_is_saved")
    return ap.parse_args()


def main():
    args = parse_args()
    qasrl_path = args.qasrl_path
    out_dir = args.out_dir

    qasrl = read_annot_csv(qasrl_path)
    ref = read_annot_csv(args.ref_path)
    sents = read_csv(args.sent_path)

    worker_data = []
    # Step 1: get a dataset for each worker
    work_groups = qasrl.groupby("worker_id")
    for worker_id, worker_df in work_groups:
        print(f"Evaluating: {worker_id}")
        # Step 2: for each worker, compare the dataset
        # with the reference on common predicate ids.
        w_pred_ids = worker_df[['qasrl_id', 'target_idx']].drop_duplicates()
        w_ref = pd.merge(ref, w_pred_ids, on=['qasrl_id', 'target_idx'])
        res = eval_datasets(worker_df, w_ref)

        # Step 3: for each worker, get argument precision and recall, and avg. number of questions per verb.
        arg_counts, larg_counts, role_counts, isnom_counts, all_matchings = res
        tp, fp, fn = arg_counts
        prec = tp / (tp + fp)
        recall = tp / (tp + fn)
        n_questions = all_matchings.sys_role.nunique()
        n_predicates = w_pred_ids.shape[0]
        qs_per_pred = float(n_questions)/n_predicates
        worker_path = os.path.join(out_dir, f"{worker_id}.csv")

        # Step 5: save report.
        all_matchings.to_csv(worker_path, index=False, encoding="utf-8")
        worker_data.append({
            "worker_id": worker_id,
            "prec": prec,
            "recall": recall,
            "n_preds": n_predicates,
            "qs_per_pred": qs_per_pred})
    worker_data = pd.DataFrame(worker_data)[['worker_id', 'n_preds', 'qs_per_pred', 'prec', 'recall']].copy()

    # Step 4: display result
    print(worker_data.sort_values(['n_preds', 'qs_per_pred'], ascending=False))


if __name__ == "__main__":
    main()


from collections import Counter
from typing import *

import pandas as pd

import qanom.evaluation.evaluate_inter_annotator as eia
from qanom import utils
from qanom.annotations.common import get_n_assignments, get_n_predicates, get_n_positive_predicates, filter_questions, \
    get_n_QAs, get_predicate_idx_label, get_n_args
from qanom.annotations.decode_encode_answers import Question, question_from_row


def isVerbalSumSeries(annot_df: pd.DataFrame) -> pd.Series:
    """
    in the returned Series is_verbal_sum, a key (= qasrl_id + _ + target_idx) is mapped to the number of generators
    for which isVerbal==True.
    """
    reduced_df = annot_df.drop_duplicates(subset=["key", "worker_id"])   # reduced has one row per predicate per worker
    is_verbal_sum = reduced_df.groupby("key").is_verbal.sum().astype(int)
    return is_verbal_sum

def isVerbalStatistics(annot_df: pd.DataFrame):
    is_verbal_sum = isVerbalSumSeries(annot_df)
    is_verbal_agreement_dist = utils.asRelative(Counter(is_verbal_sum))
    # for clarity
    utils.replaceKeys(is_verbal_agreement_dist,
                      inplace=True,
                      oldKeys2NewKeys={0: "Not Verbal",
                                       2: "Verbal",
                                       1: "Disagree"})
    return is_verbal_agreement_dist


def most_controversial_predicates(annot_df: pd.DataFrame):
    from scipy import stats
    from qanom.annotations.common import normalize, get_sent_map
    sent_map: Dict[str, List[str]] = get_sent_map(annot_df)
    cols = ['qasrl_id', 'target_idx', 'verb']
    entropies = annot_df.groupby(cols).is_verbal.agg(lambda s: stats.entropy(normalize(s.value_counts()), base=2))
    print(entropies.sort_values(ascending=False))

    #todo compete it- compress to is_verbal decision per se, and see number of annotators



"""
It may be useful to show some of the quality metrics (e.g. IAA) on different splits of the dataset, for example, for
different domains, or different kinds of questions. Here are a bunch of helper funcs to do that.
"""
def analyze_by_column_value(annot_df: pd.DataFrame,
                            column: str,
                            analysis_func: Callable[[pd.DataFrame,],Any] = eia.evaluate_inter_generator_agreement):
    for key, df in annot_df.groupby(column):
        print(f"Results for {column}=={key}:")
        res = analysis_func(df)
        print(res or "")

def analyze_by_wh_word(annot_df: pd.DataFrame,
                       analysis_func: Callable[[pd.DataFrame,],Any] = eia.evaluate_inter_generator_agreement):
    analyze_by_column_value(annot_df, 'wh', analysis_func)

def analyze_by_wh_group(annot_df: pd.DataFrame,
                        analysis_func: Callable[[pd.DataFrame,],Any] = eia.evaluate_inter_generator_agreement,
                        wh_groups={"modifiers (where, when)": ["where", "when", "how long", "how much"],
                                   "core (who, what)": ["who", "what"],
                                   "implied (how, why)": ["how", "why"]}) -> Any:
    wh2groupName = {wh: grp
                    for grp, whs in wh_groups.items()
                    for wh in whs}
    # create a new column for group, to utilize analyze_by_column_value
    grouping_column = 'wh-group'
    annot_df[grouping_column]=annot_df.wh.apply(wh2groupName.get)
    return analyze_by_column_value(annot_df, column=grouping_column, analysis_func=analysis_func)


def print_annot_statistics(annot_df: pd.DataFrame):
    n_assignments = get_n_assignments(annot_df)
    n_predicates = get_n_predicates(annot_df)
    n_positive_predicates = get_n_positive_predicates(annot_df)
    n_qas = get_n_QAs(annot_df)
    annot_with_questions_df = filter_questions(annot_df)
    roleDist = Counter(annot_with_questions_df.groupby(['key','worker_id']).agg(pd.Series.count)['question'])
    sum_roles = sum(k * v for k, v in roleDist.items())
    num_roles_average = sum_roles / float(n_assignments)
    num_positive_wo_qas = roleDist[0]
    print(f'#-predicates: {n_predicates}')
    print(f'#-positive-predicates: {n_positive_predicates} (%{100*n_positive_predicates/float(n_predicates):.1f})')
    print(f'#-QAs (total): {n_qas}')
    print(f'#-Roles per predicate Distribution: {roleDist}')
    print(f'#-Roles average (for positive predicates): {num_roles_average:.2f}')
    print(f'#-positive predicates with NO role: {num_positive_wo_qas}'
          f'  ({num_positive_wo_qas/float(n_positive_predicates):.2f}% of positives)')


def compute_num_self_cycles(annot_df: pd.DataFrame) -> int:
    """ How many arguments in annot_df include the predicate within the answer span. """
    counter = 0
    idx_lbl = get_predicate_idx_label(annot_df)
    for i,row in annot_df.iterrows():
        answers = row.answer_range
        pred_idx = row[idx_lbl]
        for a in answers:
            if pred_idx in range(a[0], a[1]):
                counter += 1
    return counter


def compute_num_self_cycles_percentage(annot_df: pd.DataFrame) -> float:
    """ Num of predicate-including args / num-args"""
    n_self_loops = compute_num_self_cycles(annot_df)
    n_args = get_n_args(annot_df)
    return float(n_self_loops) / n_args


def analyze_question_feature_distribution(annot_df: pd.DataFrame, question_func: Callable[[Question,], Any]) -> Counter:
    """ given a function f mapping Questions to any value, return the distribution of f(q) for all Q in dataset """
    values = []
    for i,row in filter_questions(annot_df).iterrows():
        question: Question = question_from_row(row)
        values.append(question_func(question))
    return Counter(values)


def question_role_distribution(annot_df: pd.DataFrame) -> Counter:
    from qanom.evaluation.roles import question_to_sem_role
    rawCounter = analyze_question_feature_distribution(annot_df, question_func=question_to_sem_role)
    return utils.asRelative(rawCounter)


from argparse import ArgumentParser
from collections import defaultdict
from itertools import combinations, permutations
from typing import Dict, Any

import pandas as pd

from qanom.annotations.common import read_annot_csv, set_key_column, get_n_predicates, get_predicate_idx_label
from qanom.annotations.decode_encode_answers import decode_qasrl
from qanom.evaluation.evaluate import eval_datasets
from qanom.evaluation.metrics import Metrics, BinaryClassificationMetrics


# describe statistics of the saved annotated data
def get_worker_statistics(annot_df: pd.DataFrame) -> Dict[str, Dict[str, Any]]:
    set_key_column(annot_df)
    by_worker = annot_df.groupby("worker_id")
    num_of_predicates = by_worker.key.nunique()
    num_of_qas = by_worker.question.count() # counts only non-NA values
    workers = list(num_of_predicates.index)
    return {worker:
                {"num-predicates": num_of_predicates[worker],
                 "num-QAs": num_of_qas[worker]}
            for worker in workers}

def get_worker_statistics_from_file(anot_fn):
    return get_worker_statistics(read_annot_csv(anot_fn))

""" Analysis functions"""


def describe_worker_work(annot_df: pd.DataFrame):
    from annotations.analyze import print_annot_statistics

    workers = annot_df.worker_id.unique().tolist()
    agreement_stats = evaluate_per_worker_iaa(annot_df, isPrinting=False)
    general_worker_stats = get_worker_statistics(annot_df)
    for worker in workers:
        """ For each worker, present:
         - # predicates
         - # of predicates judged positive (as verbal nouns)
         - # distribution of #-roles per positive predicate
        """
        print(f"********** Worker {worker}: ************")
        worker_df: pd.DataFrame = annot_df[annot_df.worker_id == worker]
        print_annot_statistics(worker_df)
        print(agreement_stats[worker])
    #  todo complete it with more info?


def evaluate_per_worker_iaa(annot_df: pd.DataFrame, isPrinting=True):
    workers = annot_df.worker_id.unique().tolist()
    n_workers = len(workers)
    n_predicates = get_n_predicates(annot_df)
    print("n_workers: ", n_workers)
    print("n_predicates: ", n_predicates)

    # save per worker a list of local 1:1 inter-annotator agreement and their weights (i.e. number of instances);
    arg_agreements = defaultdict(list)
    nom_ident_agreements = defaultdict(list)
    # go through all pairs of workers - permutations (less efficient then combinations, but more readable)
    for w1, w2 in permutations(workers, r=2):
        w1_df = annot_df[annot_df.worker_id == w1].copy()
        w2_df = annot_df[annot_df.worker_id == w2].copy()
        # compute 1:1 agreements
        arg_metrics, larg_metrics, role_metrics, nom_ident_metrics, matching_args = eval_datasets(w1_df, w2_df)
        # Arg-Accuracy: metric is f1, weight is number of predicted arguments
        arg_agreements[w1].append(arg_metrics)
        # Nominalization-Identification Accuracy: metric is accuracy, weight is number of (common) predicates
        nom_ident_agreements[w1].append(nom_ident_metrics)
    # compute per worker personal performance, measured by IAA
    worker_arg_performance = {}
    for worker_id, lst_of_agreements in arg_agreements.items():
        overallArgMetric : Metrics = sum(lst_of_agreements)
        worker_arg_performance[worker_id] = overallArgMetric
    worker_isnom_performance = {}
    for worker_id, lst_of_accMetrics in nom_ident_agreements.items():
        total_acc_metric : BinaryClassificationMetrics = sum(lst_of_accMetrics)
        worker_isnom_performance[worker_id] = total_acc_metric

    # print and save statistics
    print(f"worker_id \t\t arg \t\t\t\t is_verbal")
    worker_general_statistics = get_worker_statistics(annot_df)
    for worker_id, statistics in worker_general_statistics.items():
        if isPrinting:
            print(f"{worker_id} \t arg: {worker_arg_performance[worker_id]} " +
                  f"\t is-verbal accuracy: {worker_isnom_performance[worker_id].accuracy():.3f} " +
                  f"({statistics['num-predicates']} predicates, " +
                  f"{statistics['num-QAs']} QAs)")
        worker_general_statistics[worker_id].update({"arg-agreement": worker_arg_performance[worker_id],
                                                     "is-verbal accuracy": worker_isnom_performance[worker_id].accuracy()})
    return worker_general_statistics


def evaluate_inter_generator_agreement(annot_df: pd.DataFrame, verbose: bool = False) -> float:
    cols = ['qasrl_id', get_predicate_idx_label(annot_df)]
    n_gen = annot_df.groupby(cols).worker_id.transform(pd.Series.nunique)
    workers = annot_df.worker_id.unique().tolist()
    n_workers = len(workers)
    annot_df = annot_df.copy()
    n_predicates = annot_df[cols].drop_duplicates().shape[0]
    if verbose:
        print("n_workers: ", n_workers)
        print("n_predicates: ", n_predicates)
        print(f"metric\tworker_1\tworker_2\tprec\trecall\tf1")

    total_arg_metric = Metrics.empty()
    total_larg_metric = Metrics.empty()
    total_role_metric = Metrics.empty()
    total_nomIdent_metric : BinaryClassificationMetrics = BinaryClassificationMetrics.empty()
    for w1, w2 in combinations(workers, r=2):
        w1_df = annot_df[annot_df.worker_id == w1].copy()
        w2_df = annot_df[annot_df.worker_id == w2].copy()
        # compute agreement measures
        arg_metrics, labeled_arg_metrics, role_metrics, nom_ident_metrics, _ = \
            eval_datasets(w1_df, w2_df)
        if verbose:
            print(f"\nComparing  {w1}   to   {w2}:   [p,r,f1]")
            merged_df = pd.merge(w1_df, w2_df, on='key')
            print(f"Number of shared predicates: {get_n_predicates(merged_df)}")
            print(f"ARG:\t{arg_metrics}")
            print(f"Labeled ARG:\t{labeled_arg_metrics}")
            print(f"ROLE:\t{role_metrics}")
            print(f"NOM_IDENT:\t{w1}\t{w2}\t{nom_ident_metrics.prec():.3f}\t{nom_ident_metrics.recall():.3f}\t{nom_ident_metrics.f1():.3f}")
            print(f"NOM_IDENT accuracy: {nom_ident_metrics.accuracy():.3f}, {int(nom_ident_metrics.errors())} mismathces out of {nom_ident_metrics.instances()} predicates.")
        total_arg_metric += arg_metrics
        total_larg_metric += labeled_arg_metrics
        total_role_metric += role_metrics
        total_nomIdent_metric += nom_ident_metrics

    print(f"\nOverall pairwise agreement:")
    print(f"arg-f1 \t {total_arg_metric.f1():.4f}")
    print(f"labeled-arg-f1 \t {total_larg_metric.f1():.4f}")
    print(f"role-f1 \t {total_role_metric.f1():.4f}")
    print(f"is-verbal-accuracy \t {total_nomIdent_metric.accuracy():.4f}    for {total_nomIdent_metric.instances()} pairwise comparisons.")
    return total_arg_metric.f1()


def main(annotation_path: str):
    annot_df = read_annot_csv(annotation_path)
    annot_df = decode_qasrl(annot_df)
    # original annotations, multiple generation tasks per predicate
    print(annot_df.worker_id.value_counts())
    evaluate_inter_generator_agreement(annot_df, verbose=True)

def main_iaa_per_worker(annotation_path: str):
    annot_df = read_annot_csv(annotation_path)
    annot_df = decode_qasrl(annot_df)
    print(annot_df.worker_id.value_counts())
    evaluate_per_worker_iaa(annot_df)

if __name__ == "__main__":
    ap = ArgumentParser()
    ap.add_argument("annotation_path")
    args = ap.parse_args()
    main(args.annotation_path)





from typing import List, Dict, Tuple, Generator

import pandas as pd
from tqdm import tqdm

from qanom.annotations.common import read_annot_csv, read_csv, get_sent_map, get_predicate_idx_label
from qanom.annotations.decode_encode_answers import NO_RANGE, Argument, Role, Question, Response, decode_response
from qanom.evaluation.alignment import find_matches, align_by_argument
from qanom.evaluation.metrics import Metrics, BinaryClassificationMetrics


def to_qa_pair(roles: List[Role]) -> List[Tuple[Argument, Question]]:
    return [(arg, role.question) for role in roles for arg in role.arguments]


def build_all_qa_pairs(sys_roles: List[Role],
                       grt_roles: List[Role],
                       sys_to_grt_matches: Dict[Argument, Argument]):
    grt_qa_pairs = to_qa_pair(grt_roles)
    sys_qa_pairs = to_qa_pair(sys_roles)

    grt_qa_pairs = pd.DataFrame(grt_qa_pairs, columns=["grt_arg", "grt_role"])
    sys_qa_pairs = pd.DataFrame(sys_qa_pairs, columns=["sys_arg", "sys_role"])
    # Dictionary mapping with None values
    sys_qa_pairs['grt_arg'] = sys_qa_pairs.sys_arg.apply(sys_to_grt_matches.get)
    all_qa_pairs = pd.merge(sys_qa_pairs, grt_qa_pairs, on="grt_arg", how="outer")
    all_qa_pairs.grt_arg.fillna(NO_RANGE, inplace=True)
    all_qa_pairs.sys_arg.fillna(NO_RANGE, inplace=True)

    return all_qa_pairs


def filter_ids(df, row):
    idx_lbl = get_predicate_idx_label(df)
    return (df.qasrl_id == row.qasrl_id) & (df[idx_lbl] == row[idx_lbl])


def fill_answer(arg: Argument, tokens: List[str]):
    if arg == NO_RANGE:
        return NO_RANGE
    return " ".join(tokens[arg[0]: arg[1]])


def eval_datasets(sys_df, grt_df, sent_map= None) \
        -> Tuple[Metrics, Metrics, Metrics, BinaryClassificationMetrics, pd.DataFrame]:
    if not sent_map:
        annot_df = pd.concat([sys_df[['qasrl_id', 'sentence']], grt_df[['qasrl_id', 'sentence']]])
        sent_map = get_sent_map(annot_df)
    arg_metrics = Metrics.empty()
    labeled_arg_metrics = Metrics.empty()
    role_metrics = Metrics.empty()
    is_nom_counts = BinaryClassificationMetrics.empty()
    all_matchings = []
    for key, sys_response, grt_response in tqdm(yield_paired_predicates(sys_df, grt_df), leave=False):
        qasrl_id, target_idx = key
        tokens = sent_map[qasrl_id]
        local_arg_metric, local_labeled_arg_metric, local_role_metric, local_is_nom_metric, sys_to_grt = \
            evaluate_response(sys_response, grt_response)
        arg_metrics += local_arg_metric
        labeled_arg_metrics += local_labeled_arg_metric
        role_metrics += local_role_metric
        is_nom_counts += local_is_nom_metric
        all_args = build_all_qa_pairs(sys_response.roles, grt_response.roles, sys_to_grt)
        all_args['qasrl_id'] = qasrl_id
        all_args['target_idx'] = target_idx
        all_args['grt_arg_text'] = all_args.grt_arg.apply(fill_answer, tokens=tokens)
        all_args['sys_arg_text'] = all_args.sys_arg.apply(fill_answer, tokens=tokens)
        all_matchings.append(all_args)

    # when all_matching is empty, return empty DataFrame
    if not all_matchings:
        all_matchings = pd.DataFrame()
    else:
        all_matchings = pd.concat(all_matchings)
        all_matchings = all_matchings[['grt_arg_text', 'sys_arg_text',
                                       'grt_role', 'sys_role',
                                       'grt_arg', 'sys_arg',
                                       'qasrl_id', 'target_idx']]

    return arg_metrics, labeled_arg_metrics, role_metrics, is_nom_counts, all_matchings


def get_recall_and_precision_mistakes(sys_df, grt_df) -> (pd.DataFrame, pd.DataFrame):
    """ Return dfs of recall and precision mistakes. """
    all_matchings: pd.DataFrame = eval_datasets(sys_df, grt_df)[-1]
    # take all recall errors - roles which are NA in sys_roles in the alignment
    recall_mistakes = all_matchings[all_matchings['sys_role'].isna()]
    recall_mistakes['question'] = recall_mistakes['grt_role'].apply(lambda role: role.text)
    # take all precision errors - roles which are NA in grt_roles in the alignment
    precision_mistakes = all_matchings[all_matchings['grt_role'].isna()]
    precision_mistakes['question'] = precision_mistakes['sys_role'].apply(lambda role: role.text)
    # now produce the subsets from original DataFrames
    cols = ['qasrl_id', 'target_idx', 'question']
    recall_mistakes_df: pd.DataFrame = grt_df.merge(recall_mistakes[cols], on=cols, how="inner").drop_duplicates(cols)
    precision_mistakes_df: pd.DataFrame = sys_df.merge(precision_mistakes[cols], on=cols, how="inner").drop_duplicates(cols)

    return recall_mistakes_df, precision_mistakes_df

def eval_datasets_arg_f1(sys_df, grt_df) -> float:
    arg_metrics = eval_datasets(sys_df, grt_df)[0]
    return arg_metrics.f1()


def yield_paired_predicates(sys_df: pd.DataFrame, grt_df: pd.DataFrame) -> Generator[Tuple[Tuple[str,int],Response,Response], None, None]:
    grt_predicate_ids = grt_df[['qasrl_id', 'target_idx']].drop_duplicates()
    sys_predicate_ids = sys_df[['qasrl_id', 'target_idx']].drop_duplicates()
    # Include only those predicates which are both in grt and in sys
    predicate_ids = pd.merge(grt_predicate_ids, sys_predicate_ids, how='inner')
    for idx, row in predicate_ids.iterrows():
        sys_qa_pairs = sys_df[filter_ids(sys_df, row)].copy()
        grt_qa_pairs = grt_df[filter_ids(grt_df, row)].copy()
        sys_response = decode_response(sys_qa_pairs)
        grt_response = decode_response(grt_qa_pairs)
        yield (row.qasrl_id, row.target_idx), sys_response, grt_response


def evaluate_response(sys_response: Response,
                      grt_response: Response):
    sys_roles: List[Role] = sys_response.roles
    grt_roles: List[Role] = grt_response.roles
    sys_to_grt = find_matches(sys_response.all_args(), grt_response.all_args())

    is_nom_metrics = BinaryClassificationMetrics.simple_boolean_decision(sys_response.is_verbal, grt_response.is_verbal)

    # todo decide on evaluation of roles where is_verbal mismatch - should the roles be included in the role_count metric?
    # Currently excluding these mismatches from the arg & roles metrics
    if is_nom_metrics.errors() == 0:
        arg_metrics = eval_arguments(grt_roles, sys_roles, sys_to_grt)
        labeled_arg_metrics = eval_labeled_arguments(grt_roles, sys_roles, sys_to_grt)
        role_metrics = eval_roles(grt_roles, sys_roles, sys_to_grt)
    else:
        arg_metrics = Metrics.empty()
        labeled_arg_metrics = Metrics.empty()
        role_metrics = Metrics.empty()

    return arg_metrics, labeled_arg_metrics, role_metrics, is_nom_metrics, sys_to_grt


def count_arguments(roles: List[Role]):
    return sum(len(role.arguments) for role in roles)


def eval_arguments(grt_roles: List[Role], sys_roles: List[Role], sys_to_grt: Dict[Argument, Argument]) -> Metrics:
    tp_arg_count = len(sys_to_grt)
    fp_arg_count = count_arguments(sys_roles) - tp_arg_count
    fn_arg_count = count_arguments(grt_roles) - tp_arg_count
    return Metrics(tp_arg_count, fp_arg_count, fn_arg_count)


def eval_labeled_arguments(grt_roles: List[Role], sys_roles: List[Role], sys_to_grt: Dict[Argument, Argument]) -> Metrics:
    """ LA metric - Labeled Argument match - spans overlap and questions are equivalent. """
    tp_arg_count = count_labeled_arg_matches(grt_roles, sys_roles, sys_to_grt)
    fp_arg_count = count_arguments(sys_roles) - tp_arg_count
    fn_arg_count = count_arguments(grt_roles) - tp_arg_count
    return Metrics(tp_arg_count, fp_arg_count, fn_arg_count)


def count_labeled_arg_matches(grt_roles: List[Role], sys_roles: List[Role], sys_to_grt: Dict[Argument, Argument]) -> int:
    """ Count the number of labeled_matching among argument-match (TP) -
    i.e. only if corresponding questions are equivalent. """
    # for each argument in sys_to_grt, find corresponding question from its role
    def find_role(arg: Argument, roles: List[Role]) -> Role:
        for r in roles:
            if arg in r.arguments:
                return r

    sys_arg2q: Dict[Argument, Role] = {arg : find_role(arg, sys_roles)
                                       for arg in sys_to_grt.keys()}
    grt_arg2q: Dict[Argument, Role] = {arg : find_role(arg, grt_roles)
                                       for arg in sys_to_grt.values()}
    from qanom.evaluation.roles import is_equivalent_question

    def is_labeled_arg_match(sys_arg: Argument, grt_arg: Argument) -> bool:
        sys_question = sys_arg2q[sys_arg].question
        grt_question = grt_arg2q[grt_arg].question
        return is_equivalent_question(sys_question, grt_question)

    return sum(1
               for arg_match in sys_to_grt.items()
               if is_labeled_arg_match(*arg_match))


def eval_roles(grt_roles: List[Role],
               sys_roles: List[Role],
               sys_to_grt: Dict[Argument, Argument]) -> Metrics:
    alignemnt = align_by_argument(grt_roles, sys_roles, sys_to_grt)
    tp, fp, fn = 0, 0, 0
    for grt_role in grt_roles:
        if alignemnt.has_single_alignment(grt_role, is_grt=True):
            tp += 1
        else:
            fn += 1
    for sys_role in sys_roles:
        if not alignemnt.has_single_alignment(sys_role, is_grt=False):
            fp += 1
    return Metrics(tp, fp, fn)


def print_system_evaluation(system_df: pd.DataFrame, ground_truth_df: pd.DataFrame):
    arg, larg, role, isnom, _ = eval_datasets(system_df, ground_truth_df)

    from qanom.annotations.analyze import print_annot_statistics
    print("System prediction statistics:")
    print_annot_statistics(system_df)
    print("**************************")

    print("Ground Truth statistics:")
    print_annot_statistics(system_df)
    print("**************************")

    print("\n\t\t\t\tPrecision\tRecall\tF1")
    print(f"arg-f1 \t\t\t {arg.prec():.2f}\t{arg.recall():.2f}\t{arg.f1():.4f}")
    print(f"labeled-arg-f1 \t {larg.prec():.2f}\t{larg.recall():.2f}\t{larg.f1():.4f}")
    print(f"role-f1 \t\t {role.prec():.2f}\t{role.recall():.2f}\t{role.f1():.4f}")
    print(f"is-verbal \t\t {isnom.prec():.2f}\t{isnom.recall():.2f}\t{isnom.f1():.4f}")
    print(f"is-verbal (accuracy): \t {isnom.accuracy():.4f}    for {isnom.instances()} pairwise comparisons.")



def main(proposed_path: str, reference_path: str, sentences_path: str):
    if sentences_path:
        sent_df = read_csv(sentences_path)
        sent_map = dict(zip(sent_df.qasrl_id, sent_df.tokens.apply(str.split)))
    else:
        sent_map = None

    sys_df = read_annot_csv(proposed_path)
    grt_df = read_annot_csv(reference_path)
    print_system_evaluation(sys_df, grt_df)




"""
Reconstructing the He et. al (2015) mapping of QA-SRL questions to Roles (See Section 4 in their paper).
They define the role space (R) as:
    • R = {R0, R1, R2, R2[p], w, w[p]}
    • w isin {Where, When, Why, How, HowMuch}
    • p isin prepositions

"""
import enum
import logging

from qanom.annotations.decode_encode_answers import Question

# create logger
logger = logging.getLogger(__file__)
logger.setLevel(logging.DEBUG)

core_wh_words = {'what', 'who'}
adjunct_wh_words = {'how', 'how long', 'how much', 'when', 'where', 'why'}

all_prepositions_in_dataset = {
 'about',
 'against',
 'along',
 'around',
 'as',
 'as doing',
 'at',
 'between',
 'by',
 'by doing',
 'doing',
 'during',
 'for',
 'for doing',
 'from',
 'in',
 'in doing',
 'into',
 'of',
 'of doing',
 'on',
 'on doing',
 'out',
 'over',
 'to',
 'to do',
 'to doing',
 'towards',
 'with',
 'up to',
 'with doing'
}|{  # here O added some thta are in QANom dataset so that it will work also on QASRL-GS
     '',
     'across',
     'after',
     'ahead',
     'ahead at',
     'along with',
     'amid',
     'among',
     'aside',
     'aside for',
     'before',
     'behind',
     'do',
     'down',
     'down from',
     'from among',
     'from doing',
     'in from',
     'in to',
     'off',
     'off for',
     'off from',
     'off of',
     'off to',
     'on to',
     'on to do',
     'onto',
     'out by',
     'out of',
     'out of doing',
     'out to',
     'out to do',
     'over from',
     'through',
     'to as',
     'under',
     'up',
     'up doing',
     'up for',
     'up with',
     'upon',
     'without'}

UNGRAMMATICAL = "Ungrammatical-question" # this role is for predicted question with non-grammatical structure
"""
SemanticRole will represented by a string, denoting the roles by:
    "R0", "R1", "R2", "where", "when", ..., "R2_about", "R2_against", ..., "where_about", "where_against", ... 
"""
role_space = {"R0", "R1", "R2"} | \
             adjunct_wh_words | \
             {"R2_" + p
              for p in all_prepositions_in_dataset} | \
             {wh + "_" + p
              for wh in adjunct_wh_words
              for p in all_prepositions_in_dataset} | \
             {UNGRAMMATICAL} # this role is for predicted question with non-grammatical structure

SemanticRole = enum.Enum("SemanticRole", {r: r for r in role_space})


def is_equivalent_question(q1: Question, q2: Question) -> bool:
    return question_to_sem_role(q1) == question_to_sem_role(q2)


def question_to_sem_role(q: Question) -> SemanticRole:
    """ Refer to Table 7 at He et. al. (2015). """
    def get_R2() -> SemanticRole:
        # distinguish direct object-2 or indirect object-2 by presence of preposition
        if q.prep:
            return SemanticRole("R2_" + q.prep)
        else:
            return SemanticRole.R2

    if q.wh.lower() in core_wh_words and not q.is_passive:
        if not q.subj:
            return SemanticRole.R0
        elif not q.obj:
            return SemanticRole.R1
        elif not q.obj2 or q.obj2 in ['do', 'doing']:
            return get_R2()
        else:
            logger.warning(f"ungrammatical question: {q} ; all core args are present in core-role active question.")
            return SemanticRole(UNGRAMMATICAL)

    elif q.wh.lower() in core_wh_words and q.is_passive:
        if not q.subj:
            return SemanticRole.R1
        elif not q.obj:
            # prep=="by" is an exception
            if q.prep=="by":
                return SemanticRole.R0
            else:
                return get_R2()
        else:
            # not in their table, but can be found in data - e.g. "what was someone awarded something for?"
            if q.prep == "by":
                return SemanticRole.R0
            else:
                return get_R2()

    # voice (passive/active) has no effect on adjunct questions
    elif q.wh.lower() in adjunct_wh_words:
        if not q.obj2 and q.prep:
            return SemanticRole(q.wh + "_" + q.prep)
        else:
            return SemanticRole(q.wh)

    logger.warning(f"ungrammatical question: {q}")
    return SemanticRole(UNGRAMMATICAL)




from pathlib import Path
import os
from qanom import utils, config
heuristic_deverbal_pairs_path = Path(config.resources_path) / "verb-to-nom-heuristic" / "nom_verb_pairs.txt"

# This class documents how we built the resource
class SuffixBasedNominalizationCandidatesGenerator:
    """
    This class is only used once, to generate the `deverbal_pairs` list by exhoustively iterating the verb-seed and enumrating 
    all the optional nominalized inflections of the that verb.
    This is done by applying verb-to-nominalization morphlogical inflections hueristically. 
    We maintain this code only for scientific reproducability.
    For end users, it is enough to load and use the resulting `deverbal_pairs.txt` file.
    """
    def __init__(self) -> None:
        self.verb_seed = self.get_all_infinitives()
        self._nom_verb_pair_list = self._generate_deverbal_pairs()
        # save nom--verb pairs file 
        self._save_nom_verb_pair_list()

    def _save_nom_verb_pair_list(self):
        os.makedirs(heuristic_deverbal_pairs_path.parent, exist_ok=True)
        with open(heuristic_deverbal_pairs_path, "w") as fout:
            fout.writelines([f"{nom}\t{verb}\n" for nom, verb in self._nom_verb_pair_list])
        
    # this method is called upon import. it creates the long list of pairs and returns it
    def _generate_deverbal_pairs(self):
        """
        Return a list of (possible-nom, source-verb) based on affix heuristics for each verb in self.verb_seed.
        """
        return [(nom, verb)
                for verb in self.verb_seed
                for nom in self.verb_to_all_possible_nominalizations(verb)] 

    @classmethod
    def verb_to_all_possible_nominalizations(cls, verb):
        # returns ~17 possible nominal forms of the verb, using morphology only (some are probably not words)
        noms = [verb,
                str(cls.as_gerund(verb)),
                # -tion suffix: ("construction")
                cls.with_suffix(verb, suffix="tion", replace=["t", "te"]),
                cls.replace_suffix(verb, {"ire": "iration",
                                      "ume": "umption",
                                      "quire": "quisition",
                                      "lve": "lution"}),
                # -sion suffix: ("conclusion", "transmission")
                cls.replace_suffix(verb, {"it": "ission",
                                      "t": "sion",
                                      "de": "sion",
                                      "d": "sion",
                                      "se": "sion",
                                      "ss": "ssion"
                                      }),
                # -ment suffix: ("development")
                cls.with_suffix(verb, suffix="ment"),
                # -ance suffix: ("acceptance")
                cls.replace_suffix(verb, {"ance": "ance",
                                      "e": "ance",
                                      "": "ance",
                                      "y": "iance"}),
                # -ence suffix: ("reference")
                cls.replace_suffix(verb, {"ent": "ence",
                                      "e": "ence",
                                      "": "ence"}),
                # -ure suffix: ("failure")
                cls.with_suffix(verb, suffix="ure", replace=["e", "ure"]),
                # -al suffix: (e.g. proposal, withdrawal)
                cls.with_suffix(verb, suffix="al", replace=["e", "al"]),
               ]
        import docassemble_pattern.en as pattern
        plural_nominalizations = list(map(pattern.pluralize, set(noms)))
        all_verb_derived_noms = list(set(noms + plural_nominalizations))
        return all_verb_derived_noms

    # Lists of verbs - take union as verb-seed
    def get_all_infinitives():
        import docassemble_pattern.en as pattern

        import nltk
        from nltk.corpus import wordnet as wn
        try:
            wn.all_synsets
        except LookupError:
            print("Downloading wordnet...")
            nltk.download("wordnet")

        from nltk.corpus import verbnet
        try:  
            verbnet.lemmas  
        except LookupError:
            print("Downloading verbnet...")
            nltk.download("verbnet")
            
        # from pattern - 8.5K
        all_pattern_verbs = []
        """
        Original pattern package have bugs in this verbs.infinitives class.
        For this reason, in the annotation pipeline, we didn't use their verb seed. 
        Here we are installing pattern from the docassemble_pattern project, so bug is fixed. 
        If you wish for consistency with original QANom dataset, leave the next line commented out.
        If you care only for better coverage, uncomment the following line:
        """
        all_pattern_verbs = list(pattern.verbs.infinitives)

        # from wordnet - 8.7K verbs
        all_wn_verbs = sorted(set(l.name()
                        for v_syn in wn.all_synsets(pos="v")
                        for l in v_syn.lemmas()
                        if "_" not in l.name()))
        # from verbnet - 3.6K verbs
        all_verbnet_verbs = list(verbnet.lemmas())
        # together - 10.6K verbs
        infinitives = sorted(set(all_wn_verbs + all_pattern_verbs + all_verbnet_verbs))
        return infinitives

    def as_gerund(verb):
        import docassemble_pattern.en as pattern
        try:    # pattern's lazy-loaders have a bug (first-time failure)
            v_prog = pattern.conjugate(verb, aspect=pattern.PROGRESSIVE)
        except:
            v_prog = pattern.conjugate(verb, aspect=pattern.PROGRESSIVE)

        return v_prog
    
    @classmethod
    def with_suffix(cls, verb, suffix, replace=[]):
        for suf_to_replace in sorted(replace, key=lambda s:len(s), reverse=True):
            if verb.endswith(suf_to_replace):
                return verb.rstrip(suf_to_replace) + suffix
        return verb + suffix

    @classmethod
    def replace_suffix(cls, verb, replacements):
        # replacements is dict {"current-suffix" : "new-suffix"}
        for suf, new_suf in sorted(replacements.items(), key=lambda pair: len(pair[0]), reverse=True):
            if verb.endswith(suf):
                return verb.rstrip(suf) + new_suf
        return verb 


# This is the relevant class for users of the resource
class SuffixBasedNominalizationCandidates:

    def __init__(self):
        """
        :param verb_seed: a list of verbs (infinitive form). The seed for the artificial deverbal-nouns list.
        Generates a list of (possible-nom, source-verb) based on affix heuristics for each verb in verb_seed.
        The list is available via self.get_deverbal_pairs() method and in other formats through other methods.
        """

        # initialize the deverbal-pairs list
        self._nom_verb_pair_list = self._load_deverbal_pairs() 

        # initialize the dict
        self._affixes_based_nom_dict = self._generate_deverbal_dict(multipleEntriesPerNom=True)

    def _load_deverbal_pairs(self):
        with open(heuristic_deverbal_pairs_path) as fin:
            pairs = [line.strip().split("\t") 
                     for line in fin.readlines()]
        return pairs

    def _generate_deverbal_dict(self, multipleEntriesPerNom):
        """
        Returns:
            multipleEntriesPerNom==False:   a dict { nom : source-verb }
            multipleEntriesPerNom==True:    a dict { nom : [source-verbs] }
        Note: when multipleEntriesPerNom == False, function is not handling cases where
        the same affix-based potential nominalization is derived from multiple verbs.
        The dict will conflate these arbitrarily.
        """
        pairs = self.get_deverbal_pairs()
        if multipleEntriesPerNom:
            return utils.dictOfLists(pairs)
        else:
            return dict(pairs)


    def get_deverbal_dict(self):
        """
        Returns the field generated by self._generate_deverbal_dict()
        """
        return self._affixes_based_nom_dict


    def get_deverbal_pairs(self):
        """
        Return a list of (possible-nom, source-verb) based on affix heuristics for each verb in verb_seed.
        """
        return self._nom_verb_pair_list


    def get_all_possible_nominalizations(self):
        return self.get_deverbal_dict().keys()

    # End usage - create verb_to_nom list and query noun for source verbs
    def get_source_verbs(self, nn):
        """ Return a list of source verbs per noun if noun is in artificially generated verb_to_nom list.
            If noun is not in list, returns empty list.
        """
        if nn in self._affixes_based_nom_dict:
            return self._affixes_based_nom_dict[nn]
        else:
            return []



# A shortcut name for the class - VTN (verb-to-nominal)
VTN = SuffixBasedNominalizationCandidates


"""
Unit testing:
"""
if __name__ == '__main__':
    print("Testing SuffixBasedNominalizationCandidatesGenerator class:")
    test_verb = "show"
    all_noms = SuffixBasedNominalizationCandidatesGenerator.verb_to_all_possible_nominalizations(test_verb)
    print(f"all noms of {test_verb} are: {all_noms}\n\n")
    
    print("Testing SuffixBasedNominalizationCandidates class:")
    vtn = SuffixBasedNominalizationCandidates()
    pairs = vtn.get_deverbal_pairs()
    print(f"number of all possible_noms pairs is {len(pairs)}")
    dct = vtn.get_deverbal_dict()

    print(f'source verb for "development": {vtn.get_source_verbs("development")}')


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("qanom/version.txt", "r") as f:
    version = f.read().strip()

setuptools.setup(
    name="qanom",
    version=version,
    author="Ayal Klein",
    author_email="ayal.s.klein@gmail.com",
    description="package for Question-Answer driven Semantic Role Labeling for Nominalizations (QANom)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kleinay/QANom",
    packages=setuptools.find_packages(),
    install_requires=[
        'transformers>=2.11.0',
        'torch>=1.4',
        'nltk',
        'pandas',
        'tqdm',
        'sklearn',
    ],
    package_data={
        "": ["qanom/resources/catvar/catvar21.signed", "qanom/resources/catvar/LICENCE.txt", "qanom/resources/catvar/README.md", 
             "qanom/resources/verb-to-nom-heuristic/nom_verb_pairs.txt"],
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)


from qanom import config
import os

catvar_corpus_relative_path = "catvar/catvar21.signed" # should be in 'resources' directory
catvar_corpus_path = os.path.join(config.resources_path, catvar_corpus_relative_path)

def get_catvar_pairs():
    # catvar_pairs is a list of 16K tuples (nominalization, source-verb) for 13.8K nominalizations

    catvar_clusters = []
    with open(catvar_corpus_path, "r") as f:
        for line in f.readlines():
            cluster = [] # cluster of Words, each Word is a tuple (word-surface, pos, signature-num)
            for word_info in line.strip().split("#"):
                word_and_pos, signature = word_info.split("%")
                word, pos = word_and_pos.split("_")
                cluster.append((word, pos, int(signature)))
            catvar_clusters.append(cluster)
    def get_pairs(c):
        for noun in [w[0] for w in c if w[1]=='N']:
            for verb in [w[0] for w in c if w[1]=='V']:
                yield (noun, verb)
    catvar_pairs = [list(get_pairs(c))
                   for c in catvar_clusters
                   if len(c) > 1]
    flatten = lambda l: [item for sublist in l for item in sublist]
    catvar_pairs = flatten(catvar_pairs)
    return catvar_pairs
catvar_pairs = get_catvar_pairs()

# catvar_pairs contains 16810 pairs for 13788 nouns (keys)
def catvariate(noun):
    return [verb for n,verb in catvar_pairs if n==noun]